# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.autocomplete import products as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class AutocompleterTestCase(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

            self.engine = sa.create_engine('sqlite://')
            self.model = self.app.model
            self.model.Base.metadata.create_all(bind=self.engine)
            self.session = Session(bind=self.engine)

        def tearDown(self):
            self.session.rollback()
            self.session.close()

        def make_config(self):
            return make_config([], extend=False)


    class TestProductAutocompleter(AutocompleterTestCase):

        def make_autocompleter(self):
            return mod.ProductAutocompleter(self.config)

        def test_autocomplete(self):
            session = self.session
            model = self.model

            # first create a few products
            vinegar = model.Product(description='Apple Cider Vinegar')
            session.add(vinegar)
            dressing = model.Product(description='Apple Cider Dressing')
            session.add(dressing)
            oats = model.Product(description='Bulk Oats')
            session.add(oats)
            deleted = model.Product(description='More Oats', deleted=True)
            session.add(deleted)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'apple' yields Vinegar, Dressing
            result = self.autocompleter.autocomplete(session, 'apple')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(vinegar.uuid, uuids)
            self.assertIn(dressing.uuid, uuids)

            # search for 'oats' yields just the undeleted product
            result = self.autocompleter.autocomplete(session, 'oats')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], oats.uuid)


    class TestProductAllAutocompleter(AutocompleterTestCase):

        def make_autocompleter(self):
            return mod.ProductAllAutocompleter(self.config)

        def test_autocomplete(self):
            session = self.session
            model = self.model

            # first create a few products
            vinegar = model.Product(description='Apple Cider Vinegar')
            session.add(vinegar)
            dressing = model.Product(description='Apple Cider Dressing')
            session.add(dressing)
            oats = model.Product(description='Bulk Oats')
            session.add(oats)
            deleted = model.Product(description='More Oats', deleted=True)
            session.add(deleted)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'apple' yields Vinegar, Dressing
            result = self.autocompleter.autocomplete(session, 'apple')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(vinegar.uuid, uuids)
            self.assertIn(dressing.uuid, uuids)

            # search for 'oats' yields Bulk, More
            result = self.autocompleter.autocomplete(session, 'oats')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(oats.uuid, uuids)
            self.assertIn(deleted.uuid, uuids)


    class TestProductNewOrderAutocompleter(AutocompleterTestCase):

        def make_autocompleter(self):
            return mod.ProductNewOrderAutocompleter(self.config)

        def test_autocomplete(self):
            session = self.session
            model = self.model

            # first create a few products
            vinegar = model.Product(description='Apple Cider Vinegar',
                                    upc='074305001321')
            session.add(vinegar)
            dressing = model.Product(description='Apple Cider Dressing')
            session.add(dressing)
            oats = model.Product(description='Bulk Oats')
            session.add(oats)
            deleted = model.Product(description='More Oats', deleted=True)
            session.add(deleted)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'apple' yields Vinegar, Dressing
            result = self.autocompleter.autocomplete(session, 'apple')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(vinegar.uuid, uuids)
            self.assertIn(dressing.uuid, uuids)

            # search for unknown upc yields no results
            result = self.autocompleter.autocomplete(session, '7430500116')
            self.assertEqual(len(result), 0)

            # search for known upc yields just that product
            result = self.autocompleter.autocomplete(session, '7430500132')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], vinegar.uuid)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.autocomplete import brands as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestBrandAutocompleter(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

        def make_config(self):
            return make_config([], extend=False)

        def make_autocompleter(self):
            return mod.BrandAutocompleter(self.config)

        def test_autocomplete(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # first create a few brands
            alpha = model.Brand(name='Alpha Natural Foods')
            session.add(alpha)
            beta = model.Brand(name='Beta Natural Foods')
            session.add(beta)
            gamma = model.Brand(name='Gamma Natural Foods')
            session.add(gamma)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'natural' yields all 3 brands
            result = self.autocompleter.autocomplete(session, 'natural')
            self.assertEqual(len(result), 3)

            # search for 'gamma' yields just that brand
            result = self.autocompleter.autocomplete(session, 'gamma')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], gamma.uuid)



# -*- coding: utf-8; -*-

import os
import warnings
from unittest import TestCase

from rattail.config import make_config


class NullProgress(object):
    """
    Dummy progress bar which does nothing, but allows for better test coverage
    when used with code under test.
    """

    def __init__(self, message, count):
        pass

    def update(self, value):
        return True

    def destroy(self):
        pass

    def finish(self):
        pass


try:
    import sqlalchemy as sa
    from sqlalchemy.exc import SAWarning
    from rattail.db import model
    from rattail.db import Session
except ImportError:
    pass
else:

    warnings.filterwarnings(
        'ignore',
        r"^Dialect sqlite\+pysqlite does \*not\* support Decimal objects natively\, "
        "and SQLAlchemy must convert from floating point - rounding errors and other "
        "issues may occur\. Please consider storing Decimal numbers as strings or "
        "integers on this platform for lossless storage\.$",
        SAWarning, r'^sqlalchemy\..*$')


    class RattailMixin(object):
        """
        Generic mixin for ``TestCase`` classes which need common Rattail setup
        functionality.
        """
        engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')
        host_engine_url = os.environ.get('RATTAIL_TEST_HOST_ENGINE_URL')

        def postgresql(self):
            return self.config.rattail_engine.url.get_dialect().name == 'postgresql'

        def setUp(self):
            self.setup_rattail()

        def tearDown(self):
            self.teardown_rattail()

        def setup_rattail(self):
            config = self.make_rattail_config()
            self.config = config
            self.rattail_config = config

            engine = sa.create_engine(self.engine_url)
            config.rattail_engines = {'default': engine}
            config.rattail_engine = engine

            if self.host_engine_url:
                config.rattail_engines['host'] = sa.create_engine(self.host_engine_url)

            model = self.get_rattail_model()
            model.Base.metadata.create_all(bind=engine)

            Session.configure(bind=engine, rattail_config=config)
            self.session = Session()

        def teardown_rattail(self):
            self.session.close()
            Session.configure(bind=None, rattail_config=None)
            model = self.get_rattail_model()
            model.Base.metadata.drop_all(bind=self.rattail_config.rattail_engine)

        def make_rattail_config(self, **kwargs):
            kwargs.setdefault('files', [])
            kwargs.setdefault('use_wuttaconfig', True)
            return make_config(**kwargs)

        def get_rattail_model(self):
            return model


    class RattailTestCase(RattailMixin, TestCase):
        """
        Generic base class for Rattail tests.
        """


    class DataTestCase(TestCase):

        engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')

        def setUp(self):
            self.engine = sa.create_engine(self.engine_url)
            model.Base.metadata.create_all(bind=self.engine)
            Session.configure(bind=self.engine)
            self.session = Session()
            self.extra_setup()

        def extra_setup(self):
            """
            Derivative classes may define this as necessary, to avoid having to
            override the :meth:`setUp()` method.
            """

        def tearDown(self):
            self.session.close()
            Session.configure(bind=None)
            model.Base.metadata.drop_all(bind=self.engine)



# -*- coding: utf-8; -*-

try:
    from sqlalchemy.exc import IntegrityError
    from rattail.db import model
    # from rattail.db.changes import record_changes
    from .. import DataTestCase
except ImportError:
    pass
else:

    class SAErrorHelper(object):

        def integrity_or_flush_error(self):
            try:
                from sqlalchemy.exc import FlushError
            except ImportError:
                return IntegrityError
            else:
                return (IntegrityError, FlushError)


    class TestCustomerPerson(DataTestCase, SAErrorHelper):

        def test_customer_required(self):
            assoc = model.CustomerPerson(person=model.Person())
            self.session.add(assoc)
            self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
            self.session.rollback()
            self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
            assoc.customer = model.Customer()
            self.session.add(assoc)
            self.session.commit()
            self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

        def test_person_required(self):
            assoc = model.CustomerPerson(customer=model.Customer())
            self.session.add(assoc)
            self.assertRaises(IntegrityError, self.session.commit)
            self.session.rollback()
            self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
            assoc.person = model.Person()
            self.session.add(assoc)
            self.session.commit()
            self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

        def test_ordinal_autoincrement(self):
            customer = model.Customer()
            self.session.add(customer)
            assoc = model.CustomerPerson(person=model.Person())
            customer._people.append(assoc)
            self.session.commit()
            self.assertEqual(assoc.ordinal, 1)
            assoc = model.CustomerPerson(person=model.Person())
            customer._people.append(assoc)
            self.session.commit()
            self.assertEqual(assoc.ordinal, 2)


    class TestCustomerGroupAssignment(DataTestCase, SAErrorHelper):

        def test_customer_required(self):
            assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
            self.session.add(assignment)
            self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
            self.session.rollback()
            self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
            assignment.customer = model.Customer()
            self.session.add(assignment)
            self.session.commit()
            self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

        def test_group_required(self):
            assignment = model.CustomerGroupAssignment(customer=model.Customer())
            self.session.add(assignment)
            self.assertRaises(IntegrityError, self.session.commit)
            self.session.rollback()
            self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
            assignment.group = model.CustomerGroup()
            self.session.add(assignment)
            self.session.commit()
            self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

        def test_ordinal_autoincrement(self):
            customer = model.Customer()
            self.session.add(customer)
            assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
            customer._groups.append(assignment)
            self.session.commit()
            self.assertEqual(assignment.ordinal, 1)
            assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
            customer._groups.append(assignment)
            self.session.commit()
            self.assertEqual(assignment.ordinal, 2)


    # class TestCustomerEmailAddress(DataTestCase):

    #     def test_pop(self):
    #         customer = model.Customer()
    #         customer.add_email_address('fred.home@mailinator.com')
    #         customer.add_email_address('fred.work@mailinator.com')
    #         self.session.add(customer)
    #         self.session.commit()
    #         self.assertEqual(self.session.query(model.Customer).count(), 1)
    #         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

    #         while customer.emails:
    #             customer.emails.pop()
    #         self.session.commit()
    #         self.assertEqual(self.session.query(model.Customer).count(), 1)
    #         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

    #         # changes weren't being recorded
    #         self.assertEqual(self.session.query(model.Change).count(), 0)

    #     def test_pop_with_changes(self):
    #         record_changes(self.session)

    #         customer = model.Customer()
    #         customer.add_email_address('fred.home@mailinator.com')
    #         customer.add_email_address('fred.work@mailinator.com')
    #         self.session.add(customer)
    #         self.session.commit()
    #         self.assertEqual(self.session.query(model.Customer).count(), 1)
    #         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

    #         while customer.emails:
    #             customer.emails.pop()
    #         self.session.commit()
    #         self.assertEqual(self.session.query(model.Customer).count(), 1)
    #         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

    #         # changes should have been recorded
    #         changes = self.session.query(model.Change)
    #         self.assertEqual(changes.count(), 3)

    #         customer_change = changes.filter_by(class_name='Customer').one()
    #         self.assertEqual(customer_change.uuid, customer.uuid)
    #         self.assertFalse(customer_change.deleted)

    #         email_changes = changes.filter_by(class_name='CustomerEmailAddress')
    #         self.assertEqual(email_changes.count(), 2)
    #         self.assertEqual([x.deleted for x in email_changes], [True, True])


    class TestLabelProfile(DataTestCase):

        def test_get_printer_setting(self):
            profile = model.LabelProfile()
            self.session.add(profile)

            self.assertTrue(profile.uuid is None)
            setting = profile.get_printer_setting('some_setting')
            self.assertTrue(setting is None)
            self.assertTrue(profile.uuid is None)

            profile.uuid = 'some_uuid'
            self.session.add(model.Setting(
                    name='labels.some_uuid.printer.some_setting',
                    value='some_value'))
            self.session.flush()
            setting = profile.get_printer_setting('some_setting')
            self.assertEqual(setting, 'some_value')

        def test_save_printer_setting(self):
            self.assertEqual(self.session.query(model.Setting).count(), 0)
            profile = model.LabelProfile()
            self.session.add(profile)

            self.assertTrue(profile.uuid is None)
            profile.save_printer_setting('some_setting', 'some_value')
            self.assertFalse(profile.uuid is None)
            self.assertEqual(self.session.query(model.Setting).count(), 1)

            profile.uuid = 'some_uuid'
            profile.save_printer_setting('some_setting', 'some_value')
            self.assertEqual(self.session.query(model.Setting).count(), 2)
            setting = self.session.query(model.Setting)\
                .filter_by(name='labels.some_uuid.printer.some_setting')\
                .one()
            self.assertEqual(setting.value, 'some_value')



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock

from rattail.db import util


class TestFunctions(TestCase):

    def test_normalize_full_name(self):
        name = util.normalize_full_name(None, None)
        self.assertEqual(name, "")

        name = util.normalize_full_name("Fred", None)
        self.assertEqual(name, "Fred")

        name = util.normalize_full_name(None, "Flintstone")
        self.assertEqual(name, "Flintstone")

        name = util.normalize_full_name("Fred", "Flintstone")
        self.assertEqual(name, "Fred Flintstone")

        name = util.normalize_full_name("  Fred  ", "  Flintstone  ")
        self.assertEqual(name, "Fred Flintstone")

    def test_normalize_phone_number(self):
        number = util.normalize_phone_number(None)
        self.assertIsNone(number)

        number = util.normalize_phone_number('417-555-1234')
        self.assertEqual(number, '4175551234')

        number = util.normalize_phone_number('  (417) 555-1234  ')
        self.assertEqual(number, '4175551234')

    def test_format_phone_number(self):
        number = util.format_phone_number(None)
        self.assertIsNone(number)

        number = util.format_phone_number('417-555-1234')
        self.assertEqual(number, '(417) 555-1234')

        number = util.format_phone_number('  (417) 555-1234  ')
        self.assertEqual(number, '(417) 555-1234')


try:
    from sqlalchemy import orm
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestShortSession(TestCase):

        def test_none(self):
            with util.short_session() as s:
                self.assertIsInstance(s, Session.class_)

        def test_factory(self):
            TestSession = orm.sessionmaker()
            with util.short_session(factory=TestSession) as s:
                self.assertIsInstance(s, TestSession.class_)

        def test_Session(self):
            TestSession = orm.sessionmaker()
            with util.short_session(Session=TestSession) as s:
                self.assertIsInstance(s, TestSession.class_)

        def test_instance(self):
            # nb. nothing really happens if we provide the session instance
            session = MagicMock()
            with util.short_session(session=session) as s:
                pass
            session.commit.assert_not_called()
            session.close.assert_not_called()

        def test_config(self):
            config = MagicMock()
            TestSession = orm.sessionmaker()
            config.get_app.return_value.make_session = TestSession
            with util.short_session(config=config) as s:
                self.assertIsInstance(s, TestSession.class_)

        def test_without_commit(self):
            session = MagicMock()
            TestSession = MagicMock(return_value=session)
            with util.short_session(factory=TestSession, commit=False) as s:
                pass
            session.commit.assert_not_called()
            session.close.assert_called_once_with()

        def test_with_commit(self):
            session = MagicMock()
            TestSession = MagicMock(return_value=session)
            with util.short_session(factory=TestSession, commit=True) as s:
                pass
            session.commit.assert_called_once_with()
            session.close.assert_called_once_with()



# -*- coding: utf-8; -*-

import datetime
import warnings
from collections import OrderedDict
from unittest import TestCase

import pytest

from rattail.trainwreck import handler as mod
from rattail.config import make_config


class TestTrainwreckHandler(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.handler = self.make_handler()

    def make_config(self):
        return make_config(files=[])

    def make_handler(self):
        return mod.TrainwreckHandler(self.config)

    def test_get_trainwreck_engines(self):
        try:
            import sqlalchemy as sa
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # first let's configure 3 engines, 1 of which is hidden
        self.config.trainwreck_engines = OrderedDict([
            ('default', sa.create_engine('sqlite://')),
            ('2022', sa.create_engine('sqlite://')),
            ('2021', sa.create_engine('sqlite://')),
        ])
        self.config.setdefault('trainwreck.db', 'hide', '2022')

        # all 3 are returned by default
        engines = self.handler.get_trainwreck_engines()
        self.assertEqual(len(engines), 3)

        # but only 2 if we omit hidden
        engines = self.handler.get_trainwreck_engines(include_hidden=False)
        self.assertEqual(len(engines), 2)
        self.assertIn('default', engines)
        self.assertIn('2021', engines)
        self.assertNotIn('2022', engines)

    def test_get_hidden_engine_keys(self):
        
        # empty list returned by default
        result = self.handler.get_hidden_engine_keys()
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 0)

        # try the "legacy" setting first, to make testing simpler
        self.config.setdefault('tailbone', 'engines.trainwreck.hidden',
                               '2020, 2019, 2018')
        with warnings.catch_warnings(record=True):
            hidden = self.handler.get_hidden_engine_keys()
        self.assertEqual(hidden, ['2020', '2019', '2018'])

        # now try the "proper" setting
        self.config.setdefault('trainwreck.db', 'hide',
                               '2023, 2022, 2021')
        hidden = self.handler.get_hidden_engine_keys()
        self.assertEqual(hidden, ['2023', '2022', '2021'])

    def test_engine_is_hidden(self):
        
        # all engines are *not* hidden by default
        self.assertFalse(self.handler.engine_is_hidden('foobar'))

        # but any we explicitly hide should be reflected in call
        self.config.setdefault('trainwreck.db', 'hide',
                               '2023, 2022, 2021')
        self.assertTrue(self.handler.engine_is_hidden('2023'))
        self.assertTrue(self.handler.engine_is_hidden('2021'))
        self.assertFalse(self.handler.engine_is_hidden('2020'))
        
    def test_get_oldest_transaction_date(self):
        try:
            import sqlalchemy as sa
            from rattail.trainwreck.db import Session as TrainwreckSession
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        engine = sa.create_engine('sqlite://')
        self.config.setdefault('rattail', 'timezone.default',
                               'America/Chicago')
        self.config.setdefault('rattail.trainwreck', 'model',
                               'rattail.trainwreck.db.model.defaults')
        trainwreck = self.config.get_trainwreck_model()
        trainwreck.Base.metadata.create_all(bind=engine)
        session = TrainwreckSession(bind=engine)

        # empty db means oldest date is null
        result = self.handler.get_oldest_transaction_date(session)
        self.assertIsNone(result)

        # but if we insert a transaction, that date should be oldest
        dt = datetime.datetime(2022, 1, 1, 8)
        dt = self.handler.app.localtime(dt)
        txn = trainwreck.Transaction(end_time=dt)
        session.add(txn)
        result = self.handler.get_oldest_transaction_date(session)
        self.assertEqual(result, datetime.date(2022, 1, 1))

        # unless of course we add an older one..
        dt = datetime.datetime(2019, 6, 3, 12)
        dt = self.handler.app.localtime(dt)
        txn = trainwreck.Transaction(end_time=dt)
        session.add(txn)
        result = self.handler.get_oldest_transaction_date(session)
        self.assertEqual(result, datetime.date(2019, 6, 3))

    def test_get_newest_transaction_date(self):
        try:
            import sqlalchemy as sa
            from rattail.trainwreck.db import Session as TrainwreckSession
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        engine = sa.create_engine('sqlite://')
        self.config.setdefault('rattail', 'timezone.default',
                               'America/Chicago')
        self.config.setdefault('rattail.trainwreck', 'model',
                               'rattail.trainwreck.db.model.defaults')
        trainwreck = self.config.get_trainwreck_model()
        trainwreck.Base.metadata.create_all(bind=engine)
        session = TrainwreckSession(bind=engine)

        # empty db means newest date is null
        result = self.handler.get_newest_transaction_date(session)
        self.assertIsNone(result)

        # but if we insert a transaction, that date should be newest
        dt = datetime.datetime(2019, 6, 3, 12)
        dt = self.handler.app.localtime(dt)
        txn = trainwreck.Transaction(end_time=dt)
        session.add(txn)
        result = self.handler.get_newest_transaction_date(session)
        self.assertEqual(result, datetime.date(2019, 6, 3))

        # unless of course we add an newer one..
        dt = datetime.datetime(2022, 1, 1, 8)
        dt = self.handler.app.localtime(dt)
        txn = trainwreck.Transaction(end_time=dt)
        session.add(txn)
        result = self.handler.get_newest_transaction_date(session)
        self.assertEqual(result, datetime.date(2022, 1, 1))



# -*- coding: utf-8; -*-

import configparser
import datetime
import os
import sys
from unittest import TestCase
from unittest.mock import patch

import pytest

from wuttjamaican.testing import FileConfigTestCase
from wuttjamaican.exc import ConfigurationError

from rattail import config
from rattail.app import AppHandler


class TestRattailConfig(FileConfigTestCase):

    def test_prioritized_files(self):
        first = self.write_file('first.conf', """\
[foo]
bar = 1
""")

        second = self.write_file('second.conf', """\
[rattail.config]
require = %(here)s/first.conf
""")

        myconfig = config.RattailConfig(files=[second])
        files = myconfig.prioritized_files
        self.assertEqual(len(files), 2)
        self.assertEqual(files[0], second)
        self.assertEqual(files[1], first)

    def test_get_engine_maker(self):
        try:
            from rattail.db.config import make_engine_from_config
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # default func
        myconfig = config.RattailConfig()
        self.assertEqual(myconfig.default_engine_maker_spec, 'rattail.db.config:make_engine_from_config')
        make_engine = myconfig.get_engine_maker()
        self.assertIs(make_engine, make_engine_from_config)

    def test_setdefault(self):
        myconfig = config.RattailConfig()

        # nb. the tests below are effectively testing the custom get()
        # method in addition to setdefault()

        # value is empty by default
        self.assertIsNone(myconfig.get('foo.bar'))
        self.assertIsNone(myconfig.get('foo', 'bar'))

        # but we can change that by setting default
        myconfig.setdefault('foo.bar', 'baz')
        self.assertEqual(myconfig.get('foo.bar'), 'baz')
        self.assertEqual(myconfig.get('foo', 'bar'), 'baz')

        # also can set a default via section, option (as well as key)
        self.assertIsNone(myconfig.get('foo.blarg'))
        myconfig.setdefault('foo' ,'blarg', 'blast')
        self.assertEqual(myconfig.get('foo.blarg'), 'blast')
        self.assertEqual(myconfig.get('foo', 'blarg'), 'blast')

        # error is raised if args are ambiguous
        self.assertRaises(ValueError, myconfig.setdefault, 'foo', 'bar', 'blarg', 'blast')

        # try that for get() too
        self.assertRaises(ValueError, myconfig.get, 'foo', 'bar', 'blarg', 'blast')

    def test_get(self):
        myconfig = config.RattailConfig()
        myconfig.setdefault('foo.bar', 'baz')

        # can pass section + option
        self.assertEqual(myconfig.get('foo', 'bar'), 'baz')

        # or can pass just a key
        self.assertEqual(myconfig.get('foo.bar'), 'baz')

        # so 1 or 2 args required, otherwise error
        self.assertRaises(ValueError, myconfig.get, 'foo', 'bar', 'baz')
        self.assertRaises(ValueError, myconfig.get)

    def test_getbool(self):
        myconfig = config.RattailConfig()
        self.assertFalse(myconfig.getbool('foo.bar'))
        myconfig.setdefault('foo.bar', 'true')
        self.assertTrue(myconfig.getbool('foo.bar'))

    def test_get_date(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.get_date('foo.date'))
        myconfig.setdefault('foo.date', '2023-11-20')
        value = myconfig.get_date('foo.date')
        self.assertIsInstance(value, datetime.date)
        self.assertEqual(value, datetime.date(2023, 11, 20))

    def test_getdate(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getdate('foo.date'))
        myconfig.setdefault('foo.date', '2023-11-20')
        value = myconfig.getdate('foo.date')
        self.assertIsInstance(value, datetime.date)
        self.assertEqual(value, datetime.date(2023, 11, 20))

    def test_getint(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getint('foo.bar'))
        myconfig.setdefault('foo.bar', '42')
        self.assertEqual(myconfig.getint('foo.bar'), 42)

    def test_getlist(self):
        myconfig = config.RattailConfig()
        self.assertIsNone(myconfig.getlist('foo.bar'))
        myconfig.setdefault('foo.bar', 'hello world')
        self.assertEqual(myconfig.getlist('foo.bar'), ['hello', 'world'])

    def test_parse_bool(self):
        myconfig = config.RattailConfig()
        self.assertTrue(myconfig.parse_bool('true'))
        self.assertFalse(myconfig.parse_bool('false'))

    def test_parse_list(self):
        myconfig = config.RattailConfig()
        self.assertEqual(myconfig.parse_list(None), [])
        self.assertEqual(myconfig.parse_list('hello world'), ['hello', 'world'])

    def test_make_list_string(self):
        myconfig = config.RattailConfig()

        value = myconfig.make_list_string(['foo', 'bar'])
        self.assertEqual(value, 'foo, bar')

        value = myconfig.make_list_string(['hello world', 'how are you'])
        self.assertEqual(value, "'hello world', 'how are you'")

        value = myconfig.make_list_string(["you don't", 'say'])
        self.assertEqual(value, "\"you don't\", say")

    def test_get_app(self):
        myconfig = config.RattailConfig()
        app = myconfig.get_app()
        self.assertIsInstance(app, AppHandler)
        self.assertIs(type(app), AppHandler)

    def test_beaker_invalidate_setting(self):
        # TODO: this doesn't really test anything, just gives coverage
        myconfig = config.RattailConfig()
        myconfig.beaker_invalidate_setting('foo')

    def test_node_type(self):
        myconfig = config.RattailConfig()

        # error if node type not defined
        self.assertRaises(ConfigurationError, myconfig.node_type)

        # unless default is provided
        self.assertEqual(myconfig.node_type(default='foo'), 'foo')

        # or config contains the definition
        myconfig.setdefault('rattail.node_type', 'bar')
        self.assertEqual(myconfig.node_type(), 'bar')

    def test_get_model(self):
        try:
            import sqlalchemy
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # default is rattail.db.model
        myconfig = config.RattailConfig()
        model = myconfig.get_model()
        self.assertIs(model, sys.modules['rattail.db.model'])

        # or config may specify
        myconfig = config.RattailConfig()
        myconfig.setdefault('rattail.model', 'rattail.trainwreck.db.model')
        model = myconfig.get_model()
        self.assertIs(model, sys.modules['rattail.trainwreck.db.model'])

    def test_get_enum(self):
        myconfig = config.RattailConfig()

        # default is rattail.enum
        enum = myconfig.get_enum()
        self.assertIs(enum, sys.modules['rattail.enum'])

        # or config may specify
        # (nb. using bogus example module here)
        myconfig.setdefault('rattail.enum', 'rattail.util')
        enum = myconfig.get_enum()
        self.assertIs(enum, sys.modules['rattail.util'])

    def test_get_trainwreck_model(self):
        try:
            import sqlalchemy
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        myconfig = config.RattailConfig()

        # error if not defined
        self.assertRaises(ConfigurationError, myconfig.get_trainwreck_model)

        # but config may specify
        myconfig.setdefault('rattail.trainwreck.model', 'rattail.trainwreck.db.model')
        model = myconfig.get_trainwreck_model()
        self.assertIs(model, sys.modules['rattail.trainwreck.db.model'])

    def test_versioning_enabled(self):
        myconfig = config.RattailConfig()

        # false by default
        self.assertFalse(myconfig.versioning_enabled())

        # but config may enable
        myconfig.setdefault('rattail.db.versioning.enabled', 'true')
        self.assertTrue(myconfig.versioning_enabled())

    def test_app_package(self):
        myconfig = config.RattailConfig()

        # error if not defined
        self.assertRaises(ConfigurationError, myconfig.app_package)

        # unless default is provided
        self.assertEqual(myconfig.app_package(default='foo'), 'foo')

        # but config may specify
        myconfig.setdefault('rattail.app_package', 'bar')
        self.assertEqual(myconfig.app_package(), 'bar')

    def test_app_title(self):
        myconfig = config.RattailConfig()

        # default title
        self.assertEqual(myconfig.app_title(), 'Rattail')

        # but config may specify
        myconfig.setdefault('rattail.app_title', 'Foo')
        self.assertEqual(myconfig.app_title(), 'Foo')

    def test_node_title(self):
        myconfig = config.RattailConfig()

        # default title
        self.assertEqual(myconfig.app_title(), 'Rattail')

        # but config may specify
        myconfig.setdefault('rattail.node_title', 'Foo (node)')
        self.assertEqual(myconfig.node_title(), 'Foo (node)')

    def test_running_from_source(self):
        myconfig = config.RattailConfig()

        # false by default
        self.assertFalse(myconfig.running_from_source())

        # but config may enable
        myconfig.setdefault('rattail.running_from_source', 'true')
        self.assertTrue(myconfig.running_from_source())

    def test_demo(self):
        myconfig = config.RattailConfig()

        # false by default
        self.assertFalse(myconfig.demo())

        # but config may enable
        myconfig.setdefault('rattail.demo', 'true')
        self.assertTrue(myconfig.demo())

    def test_appdir(self):
        myconfig = config.RattailConfig()

        # can be none if required is false
        self.assertIsNone(myconfig.appdir(require=False))

        # otherwise sane fallback is used
        with patch('rattail.config.sys') as sys:
            sys.prefix = 'foo'
            path = os.path.join('foo', 'app')
            self.assertEqual(myconfig.appdir(), path)

        # or config may specify
        myconfig.setdefault('rattail.appdir', '/foo/bar/baz')
        self.assertEqual(myconfig.appdir(), '/foo/bar/baz')

    def test_datadir(self):
        myconfig = config.RattailConfig()

        # error if not defined
        self.assertRaises(ConfigurationError, myconfig.datadir)

        # but can avoid error if not required
        self.assertIsNone(myconfig.datadir(require=False))

        # or config may specify
        myconfig.setdefault('rattail.datadir', '/foo/bar/baz')
        self.assertEqual(myconfig.datadir(), '/foo/bar/baz')

    def test_workdir(self):
        myconfig = config.RattailConfig()

        # error if not defined
        self.assertRaises(ConfigurationError, myconfig.workdir)

        # but can avoid error if not required
        self.assertIsNone(myconfig.workdir(require=False))

        # or config may specify
        myconfig.setdefault('rattail.workdir', '/foo/bar/baz')
        self.assertEqual(myconfig.workdir(), '/foo/bar/baz')

    def test_batch_filedir(self):
        myconfig = config.RattailConfig()

        # error if not defined
        self.assertRaises(ConfigurationError, myconfig.batch_filedir)

        # config may specify
        path = os.path.join(os.sep, 'foo', 'files')
        myconfig.setdefault('rattail.batch.files', path)
        self.assertEqual(myconfig.batch_filedir(), path)

        # caller may specify a key
        self.assertEqual(myconfig.batch_filedir(key='bar'), os.path.join(path, 'bar'))


class TestRattailDefaultFiles(FileConfigTestCase):

    def test_quiet_conf(self):
        generic = self.write_file('generic.conf', '')
        quiet = self.write_file('quiet.conf', '')

        with patch('rattail.config.generic_default_files') as generic_default_files:
            generic_default_files.return_value = [generic]

            with patch('rattail.config.os') as mockos:
                mockos.path.join.return_value = quiet

                # generic files by default
                mockos.path.exists.return_value = False
                files = config.rattail_default_files('rattail')
                generic_default_files.assert_called_once_with('rattail')
                self.assertEqual(files, [generic])

                # but if quiet.conf exists, will return that
                generic_default_files.reset_mock()
                mockos.path.exists.return_value = True
                files = config.rattail_default_files('rattail')
                generic_default_files.assert_not_called()
                self.assertEqual(files, [quiet])


class TestMakeConfig(FileConfigTestCase):

    def test_files(self):
        generic = self.write_file('generic.conf', '')
        myfile = self.write_file('my.conf', '')

        # generic files by default
        myconfig = config.make_config(default_files=[generic])
        self.assertEqual(myconfig.files_read, [generic])

        # can specify single primary file
        myconfig = config.make_config(myfile, default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])

        # can specify primary files as list
        myconfig = config.make_config([myfile], default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])

        # can specify primary files via env
        myconfig = config.make_config(env={'RATTAIL_CONFIG_FILES': myfile},
                                      default_files=[generic])
        self.assertEqual(myconfig.files_read, [myfile])



# -*- coding: utf-8; -*-
"""
Alembic Environment Script
"""

from alembic import context

from rattail.config import make_config


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
alembic_config = context.config

# Use same config file for Rattail, as we are for Alembic.
rattail_config = make_config(alembic_config.config_file_name,
                             usedb=False,
                             versioning=False)

# add your model's MetaData object here
# for 'autogenerate' support
model = rattail_config.get_trainwreck_model()
target_metadata = model.Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    engine = rattail_config.trainwreck_engine
    context.configure(
        url=engine.url,
        target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = rattail_config.trainwreck_engine
    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Configuration
"""

import logging
import warnings

from wuttjamaican.util import parse_bool
from wuttjamaican.db.conf import make_engine_from_config as wutta_make_engine_from_config

from rattail.exceptions import SQLAlchemyNotInstalled


log = logging.getLogger(__name__)


def get_engines(config, section='rattail.db'):
    """
    DEPRECATED; please use
    :func:`wuttjamaican:wuttjamaican.db.conf.get_engines()` instead.
    """
    warnings.warn("rattail.db.config.get_engines() is deprecated; "
                  "please use wuttjamaican.db.get_engines() instead",
                  DeprecationWarning, stacklevel=2)
    from wuttjamaican.db import get_engines
    return get_engines(config, section)


def get_default_engine(config, section='rattail.db'):
    """
    DEPRECATED; please use
    :func:`wuttjamaican:wuttjamaican.db.conf.get_engines()` instead.
    """
    warnings.warn("rattail.db.config.get_default_engine() is deprecated; "
                  "please use wuttjamaican.db.get_engines() instead",
                  DeprecationWarning, stacklevel=2)
    from wuttjamaican.db import get_engines
    return get_engines(config, section).get('default')


# TODO: Deprecate/remove this.
def configure_session(config, session):
    """
    Configure a session factory or instance.  Currently all this does is
    install the hook to record changes, if config so dictates.
    """
    if config.getbool('rattail.db', 'changes.record', usedb=False):
        from rattail.db.changes import record_changes
        record_changes(session, config=config)


def configure_versioning(config, force=False, manager=None, plugins=None, **kwargs):
    """
    Configure Continuum versioning.
    """
    if not config.versioning_enabled() and not force:
        return

    try:
        from sqlalchemy.orm import configure_mappers
        import sqlalchemy_continuum as continuum
        from sqlalchemy_continuum.plugins import TransactionMetaPlugin
        from rattail.db.continuum import versioning_manager, RattailPlugin
    except ImportError as error:
        raise SQLAlchemyNotInstalled(error)
    else:
        kwargs['manager'] = manager or versioning_manager
        if plugins:
            kwargs['plugins'] = plugins
        else:
            kwargs['plugins'] = [TransactionMetaPlugin(), RattailPlugin()]
        log.debug("enabling Continuum versioning")
        continuum.make_versioned(**kwargs)

        # TODO: is this the best way/place to confirm versioning?
        app = config.get_app()
        try:
            model = app.model
            configure_mappers()
            transaction_class = continuum.transaction_class(model.User)
            config.versioning_has_been_enabled = True
        except continuum.ClassNotVersioned:
            raise RuntimeError("Versioning is enabled and configured, but is not functional!  "
                               "This probably means the code import sequence is faulty somehow.  "
                               "Please investigate ASAP.")


def make_engine_from_config(
        config_dict,
        prefix='sqlalchemy.',
        **kwargs):
    """
    This is the same as
    :func:`wuttjamaican:wuttjamaican.db.conf.make_engine_from_config()`
    except Rattail may customize the engine a bit further:

    The engine can be told to "record changes" for sake of
    datasync; for instance:

    .. code-block:: ini

       [rattail.db]
       default.url = sqlite:///
       default.record_changes = true

    And/or the engine can be told to log its SQLAlchemy connection
    pool status:

    .. code-block:: ini

       [rattail.db]
       default.url = sqlite:///
       default.log_pool_status = true
    """
    config_dict = dict(config_dict)

    # stash flag for recording changes
    record_changes = False
    key = f'{prefix}record_changes'
    if key in config_dict:
        record_changes = parse_bool(config_dict.pop(key))

    # stash flag for logging db pool status
    log_pool_status = False
    key = f'{prefix}log_pool_status'
    if key in config_dict:
        log_pool_status = parse_bool(config_dict.pop(key))

    # make engine per usual
    engine = wutta_make_engine_from_config(config_dict, prefix=prefix, **kwargs)

    # then apply flags from stash
    if record_changes:
        engine.rattail_record_changes = True
    if log_pool_status:
        engine.rattail_log_pool_status = log_pool_status

    return engine



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import MagicMock, patch

from wuttjamaican.testing import FileConfigTestCase

from rattail.commands import typer as mod


class TestMakeCliConfig(FileConfigTestCase):

    def test_basic(self):
        # nb. must specify config file to avoid any files in dev environment
        myconf = self.write_file('my.conf', '')
        ctx = MagicMock(params={
            'config_paths': [myconf],
        })
        config = mod.make_cli_config(ctx)
        self.assertEqual(config.files_read, [myconf])

    def test_no_init(self):
        ctx = MagicMock(params={
            'no_init': True,
        })
        config = mod.make_cli_config(ctx)
        self.assertEqual(config.files_read, [])

    def test_bad_model(self):
        myconf = self.write_file('my.conf', """
[rattail]
model_spec = invalid_model_spec
""")
        ctx = MagicMock(params={
            'config_paths': [myconf],
        })
        config = mod.make_cli_config(ctx)
        app = config.get_app()
        self.assertRaises(ImportError, getattr, app, 'model')

    def test_versioning(self):
        # nb. must specify config file to avoid any files in dev environment
        myconf = self.write_file('my.conf', '')
        ctx = MagicMock(params={
            'config_paths': [myconf],
            'versioning': True,
        })

        configure_versioning = MagicMock()
        mock_config = MagicMock(configure_versioning=configure_versioning)
        with patch.dict('sys.modules', **{'rattail.db.config': mock_config}):
            config = mod.make_cli_config(ctx)
            configure_versioning.assert_called_once()

    def test_missing_sqlalchemy(self):
        # nb. must specify config file to avoid any files in dev environment
        myconf = self.write_file('my.conf', '')
        ctx = MagicMock(params={
            'config_paths': [myconf],
        })

        orig_import = __import__

        def mock_import(name, globals=None, locals=None, fromlist={}, level=0):
            if name == 'rattail.db.config' and fromlist == ('configure_versioning',):
                return ImportError
            return orig_import(name, globals=globals, locals=locals, fromlist=fromlist, level=level)

        with patch('builtins.__import__', side_effect=mock_import):
            # nb. just make sure there are no errors
            config = mod.make_cli_config(ctx)



# -*- coding: utf-8 -*-
"""add label/handheld batch(es) tie

Revision ID: 3e79dd89fe12
Revises: 043d4bafc0be
Create Date: 2017-06-21 14:19:14.308813

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3e79dd89fe12'
down_revision = u'043d4bafc0be'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql
from wuttjamaican.util import make_uuid


def upgrade():

    # label_batch_handheld
    label_batch_handheld = op.create_table('label_batch_handheld',
                                           sa.Column('uuid', sa.String(length=32), nullable=False),
                                           sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                                           sa.Column('handheld_uuid', sa.String(length=32), nullable=False),
                                           sa.Column('ordinal', sa.Integer(), nullable=False),
                                           sa.ForeignKeyConstraint(['batch_uuid'], [u'label_batch.uuid'], name=u'label_batch_handheld_fk_batch'),
                                           sa.ForeignKeyConstraint(['handheld_uuid'], [u'batch_handheld.uuid'], name=u'label_batch_handheld_fk_handheld'),
                                           sa.PrimaryKeyConstraint('uuid')
    )

    # migrate data from label batch header, to new association table
    label_batch = sa.sql.table('label_batch',
                               sa.sql.column('uuid'),
                               sa.sql.column('handheld_batch_uuid'))
    cursor = op.get_bind().execute(label_batch.select())
    for row in cursor.fetchall():
        if row['handheld_batch_uuid']:
            op.get_bind().execute(label_batch_handheld.insert().values({
                'uuid': make_uuid(),
                'batch_uuid': row['uuid'],
                'handheld_uuid': row['handheld_batch_uuid'],
                'ordinal': 1,
            }))


def downgrade():

    # label_batch_handheld
    op.drop_table('label_batch_handheld')



# -*- coding: utf-8; -*-

import os
from collections import OrderedDict
from unittest import TestCase
from unittest.mock import Mock, patch, call

from rattail.importing import importers
from rattail.config import make_config, RattailConfig
from .. import NullProgress
from . import ImporterTester


class ImporterBattery(ImporterTester):
    """
    Battery of tests which can hopefully be ran for any non-bulk importer.
    """

    def setUp(self):
        self.config = self.make_config()

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Los_Angeles')
        return config

    def test_import_data_empty(self):
        importer = self.make_importer()
        result = importer.import_data()
        self.assertEqual(result, {})

    def test_import_data_dry_run(self):
        importer = self.make_importer()
        self.assertFalse(importer.dry_run)
        importer.import_data(dry_run=True)
        self.assertTrue(importer.dry_run)

    def test_import_data_create(self):
        importer = self.make_importer()
        with patch.object(importer, 'get_key', lambda k: k):
            with patch.object(importer, 'create_object') as create:
                importer.import_data(host_data=[1, 2, 3])
                self.assertEqual(create.call_args_list, [
                    call(1, 1), call(2, 2), call(3, 3)])

    def test_import_data_max_create(self):
        importer = self.make_importer()
        with patch.object(importer, 'get_key', lambda k: k):
            with patch.object(importer, 'create_object') as create:
                importer.import_data(host_data=[1, 2, 3], max_create=1)
                self.assertEqual(create.call_args_list, [call(1, 1)])


class BulkImporterBattery(ImporterBattery):
    """
    Battery of tests which can hopefully be ran for any bulk importer.
    """

    def test_import_data_empty(self):
        importer = self.make_importer()
        result = importer.import_data()
        self.assertEqual(result, 0)


class TestImporter(TestCase):

    def setUp(self):
        self.config = self.make_config()

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Los_Angeles')
        return config

    def test_init(self):

        # defaults
        importer = importers.Importer(self.config)
        self.assertIsNone(importer.model_class)
        self.assertIsNone(importer.key)
        self.assertEqual(importer.fields, [])
        self.assertIsNone(importer.host_system_title)

        # key must be included among the fields
        self.assertRaises(ValueError, importers.Importer, key='upc', fields=[])
        importer = importers.Importer(self.config, key='upc', fields=['upc'])
        self.assertEqual(importer.key, ('upc',))
        self.assertEqual(importer.fields, ['upc'])

        # extra bits are passed as-is
        importer = importers.Importer(self.config)
        self.assertFalse(hasattr(importer, 'extra_bit'))
        extra_bit = object()
        importer = importers.Importer(self.config, extra_bit=extra_bit)
        self.assertIs(importer.extra_bit, extra_bit)

    def test_delete_flag(self):
        # disabled by default
        importer = importers.Importer(self.config)
        self.assertTrue(importer.allow_delete)
        self.assertFalse(importer.delete)
        importer.import_data(host_data=[])
        self.assertFalse(importer.delete)

        # but can be enabled
        importer = importers.Importer(self.config, delete=True)
        self.assertTrue(importer.allow_delete)
        self.assertTrue(importer.delete)
        importer = importers.Importer(self.config)
        self.assertFalse(importer.delete)
        importer.import_data(host_data=[], delete=True)
        self.assertTrue(importer.delete)

    def test_get_host_objects(self):
        importer = importers.Importer(self.config)
        objects = importer.get_host_objects()
        self.assertEqual(objects, [])

    def test_cache_local_data(self):
        importer = importers.Importer(self.config)
        self.assertRaises(NotImplementedError, importer.cache_local_data)

    def test_get_local_object(self):
        importer = importers.Importer(self.config)
        self.assertFalse(importer.caches_local_data)
        self.assertRaises(NotImplementedError, importer.get_local_object, None)

        someobj = object()
        with patch.object(importer, 'get_single_local_object', Mock(return_value=someobj)):
            obj = importer.get_local_object('somekey')
        self.assertIs(obj, someobj)

        importer.caches_local_data = True
        importer.cached_local_data = {'somekey': {'object': someobj, 'data': {}}}
        obj = importer.get_local_object('somekey')
        self.assertIs(obj, someobj)

    def test_get_single_local_object(self):
        importer = importers.Importer(self.config)
        self.assertRaises(NotImplementedError, importer.get_single_local_object, None)

    def test_get_cache_key(self):
        importer = importers.Importer(self.config, key='upc', fields=['upc'])
        obj = {'upc': '00074305001321'}
        normal = {'data': obj}
        key = importer.get_cache_key(obj, normal)
        self.assertEqual(key, ('00074305001321',))

    def test_normalize_cache_object(self):
        importer = importers.Importer(self.config)
        obj = {'upc': '00074305001321'}
        with patch.object(importer, 'normalize_local_object', new=lambda obj: obj):
            cached = importer.normalize_cache_object(obj)
        self.assertEqual(cached, {'object': obj, 'data': obj})

    def test_normalize_local_object(self):
        importer = importers.Importer(self.config, key='upc',
                                      fields=['upc', 'description'])
        importer.simple_fields = importer.fields
        obj = Mock(upc='00074305001321', description="Apple Cider Vinegar")
        data = importer.normalize_local_object(obj)
        self.assertEqual(data, {'upc': '00074305001321', 'description': "Apple Cider Vinegar"})

    def test_update_object(self):
        importer = importers.Importer(self.config, key='upc',
                                      fields=['upc', 'description'])
        importer.simple_fields = importer.fields
        obj = Mock(upc='00074305001321', description="Apple Cider Vinegar")

        newobj = importer.update_object(obj, {'upc': '00074305001321', 'description': "Apple Cider Vinegar"})
        self.assertIs(newobj, obj)
        self.assertEqual(obj.description, "Apple Cider Vinegar")

        newobj = importer.update_object(obj, {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"})
        self.assertIs(newobj, obj)
        self.assertEqual(obj.description, "Apple Cider Vinegar 32oz")

    def test_normalize_host_data(self):
        importer = importers.Importer(self.config, key='upc',
                                      fields=['upc', 'description'],
                                      progress=NullProgress)

        data = [
            {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"},
            {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"},
        ]

        host_data = importer.normalize_host_data(host_objects=[])
        self.assertEqual(host_data, [])

        host_data = importer.normalize_host_data(host_objects=data)
        self.assertEqual(host_data, data)

        with patch.object(importer, 'get_host_objects', new=Mock(return_value=data)):
            host_data = importer.normalize_host_data()
        self.assertEqual(host_data, data)

    def test_get_deletion_keys(self):
        importer = importers.Importer(self.config)
        self.assertFalse(importer.caches_local_data)
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set())

        importer.caches_local_data = True
        self.assertIsNone(importer.cached_local_data)
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set())

        importer.cached_local_data = {
            'delete-me': {
                'object': object(),
                'data': {},
            },
        }
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set(['delete-me']))

    def test_model_name_attr(self):
        # default is None
        importer = importers.Importer(self.config)
        self.assertIsNone(importer.model_name)

        # but may be overridden via init kwarg
        importer = importers.Importer(self.config, model_name='Foo')
        self.assertEqual(importer.model_name, 'Foo')

        # or may inherit its value from 'model_class'
        class Foo:
            pass
        importer = importers.Importer(self.config, model_class=Foo)
        self.assertEqual(importer.model_name, 'Foo')

    def test_batch_size_attr(self):
        # default is 200
        importer = importers.Importer(self.config)
        self.assertEqual(importer.batch_size, 200)

        # but may be overridden via init kwarg
        importer = importers.Importer(self.config, batch_size=0)
        self.assertEqual(importer.batch_size, 0)
        importer = importers.Importer(self.config, batch_size=42)
        self.assertEqual(importer.batch_size, 42)

        # batch size determines how often flush occurs
        data = [{'id': i} for i in range(1, 101)]
        importer = importers.Importer(self.config, model_name='Foo',
                                      key='id', fields=['id'],
                                      empty_local_data=True)
        importer.handler = Mock(local_title="Nevermind")
        with patch.object(importer, 'create_object'): # just mock that out
            with patch.object(importer, 'flush_create_update') as flush:

                # 4 batches @ 33/per
                importer.import_data(host_data=data, batch_size=33)
                self.assertEqual(flush.call_count, 4)
                flush.reset_mock()

                # 3 batches @ 34/per
                importer.import_data(host_data=data, batch_size=34)
                self.assertEqual(flush.call_count, 3)
                flush.reset_mock()

                # 2 batches @ 50/per
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # one extra/final flush happens, whenever the total number of
                # changes happens to match the batch size...

                # 1 batch @ 100/per, plus final flush
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # 1 batch @ 200/per
                importer.import_data(host_data=data, batch_size=200)
                self.assertEqual(flush.call_count, 1)
                flush.reset_mock()

                # one extra/final flush also happens when batching is disabled

                # 100 "batches" @ 0/per, plus final flush
                importer.import_data(host_data=data, batch_size=0)
                self.assertEqual(flush.call_count, 101)
                flush.reset_mock()


class TestBulkImporter(BulkImporterBattery, TestCase):
    importer_class = importers.BulkImporter
        
    def test_batch_size_attr(self):
        # default is 200
        importer = importers.BulkImporter(self.config)
        self.assertEqual(importer.batch_size, 200)

        # but may be overridden via init kwarg
        importer = importers.BulkImporter(self.config, batch_size=0)
        self.assertEqual(importer.batch_size, 0)
        importer = importers.BulkImporter(self.config, batch_size=42)
        self.assertEqual(importer.batch_size, 42)

        # batch size determines how often flush occurs
        data = [{'id': i} for i in range(1, 101)]
        importer = importers.BulkImporter(self.config, model_name='Foo',
                                          key='id', fields=['id'],
                                          empty_local_data=True)
        with patch.object(importer, 'create_object'): # just mock that out
            with patch.object(importer, 'flush_create_update') as flush:

                # 4 batches @ 33/per
                importer.import_data(host_data=data, batch_size=33)
                self.assertEqual(flush.call_count, 4)
                flush.reset_mock()

                # 3 batches @ 34/per
                importer.import_data(host_data=data, batch_size=34)
                self.assertEqual(flush.call_count, 3)
                flush.reset_mock()

                # 2 batches @ 50/per
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # one extra/final flush happens, whenever the total number of
                # changes happens to match the batch size...

                # 1 batch @ 100/per, plus final flush
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # 1 batch @ 200/per
                importer.import_data(host_data=data, batch_size=200)
                self.assertEqual(flush.call_count, 1)
                flush.reset_mock()

                # one extra/final flush also happens when batching is disabled

                # 100 "batches" @ 0/per, plus final flush
                importer.import_data(host_data=data, batch_size=0)
                self.assertEqual(flush.call_count, 101)
                flush.reset_mock()


######################################################################
# fake importer class, tested mostly for basic coverage
######################################################################

class Product(object):
    upc = None
    description = None


class MockImporter(importers.Importer):
    model_class = Product
    key = 'upc'
    simple_fields = ['upc', 'description']
    supported_fields = simple_fields
    caches_local_data = True
    flush_every_x = 1
    session = Mock()

    def normalize_local_object(self, obj):
        return obj

    def update_object(self, obj, host_data, local_data=None):
        return host_data


class TestMockImporter(ImporterTester, TestCase):
    importer_class = MockImporter

    sample_data = OrderedDict([
        ('16oz', {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"}),
        ('32oz', {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"}),
        ('1gal', {'upc': '00074305011283', 'description': "Apple Cider Vinegar 1gal"}),
    ])

    def setUp(self):
        self.config = self.make_config()
        self.importer = self.make_importer()
        self.importer.handler = Mock(local_title="Nevermind")

    def make_config(self):
        config = make_config([], extend=False)
        config.setdefault('rattail', 'timezone.default', 'America/Los_Angeles')
        return config

    def test_create(self):
        local = self.copy_data()
        del local['32oz']
        self.import_data(local_data=local)
        self.assert_import_created('32oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_create_empty(self):
        self.import_data(host_data={}, local_data={})
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong description"
        self.import_data(local_data=local)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_delete(self):
        local = self.copy_data()
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus')

    def test_duplicate(self):
        host = self.copy_data()
        host['32oz-dupe'] = host['32oz']
        self.import_data(host_data=host)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_create=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_total_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_update=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_total_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_delete=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_total=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete_with_changes(self):
        local = self.copy_data()
        del local['16oz']
        local['32oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_total=3)
        self.assert_import_created('16oz')
        self.assert_import_updated('32oz', '1gal')
        self.assert_import_deleted()


try:
    import sqlalchemy as sa
    from rattail.db import model, Session
    from rattail.db.util import QuerySequence
except ImportError:
    pass
else:

    class TestFromQuery(TestCase):

        def setUp(self):
            self.config = RattailConfig(defaults={
                'rattail.timezone.default': 'America/Chicago',
            })
            engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')
            self.engine = sa.create_engine(engine_url)
            model.Base.metadata.create_all(bind=self.engine)
            Session.configure(bind=self.engine)

        def tearDown(self):
            model.Base.metadata.drop_all(bind=self.engine)
            Session.configure(bind=None)

        def test_query(self):
            importer = importers.FromQuery(self.config)
            self.assertRaises(NotImplementedError, importer.query)

        def test_get_host_objects(self):
            app = self.config.get_app()
            session = app.make_session()
            query = session.query(model.Product)
            importer = importers.FromQuery(self.config)
            with patch.object(importer, 'query', Mock(return_value=query)):
                objects = importer.get_host_objects()
            self.assertIsInstance(objects, QuerySequence)
            session.close()



# -*- coding: utf-8; -*-

from unittest import TestCase

try:
    from sqlalchemy import create_engine
    from sqlalchemy import Column, Integer, String, ForeignKey
    from sqlalchemy.orm import sessionmaker, relationship
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.ext.associationproxy import association_proxy
    from rattail.db import core
except ImportError:
    pass
else:

    class TestGetSetFactory(TestCase):

        def setUp(self):
            Base = declarative_base()

            class Primary(Base):
                __tablename__ = 'primary'
                id = Column(Integer(), primary_key=True)
                foo = Column(String(length=10))

            class Secondary(Base):
                __tablename__ = 'secondary'
                id = Column(Integer(), primary_key=True)
                primary_id = Column(Integer(), ForeignKey('primary.id'))
                bar = Column(String(length=10))

            Primary._secondary = relationship(
                Secondary, backref='primary', uselist=False)
            Primary.bar = association_proxy(
                '_secondary', 'bar',
                getset_factory=core.getset_factory)

            self.Primary = Primary
            self.Secondary = Secondary

            engine = create_engine('sqlite://')
            Base.metadata.create_all(bind=engine)
            Session = sessionmaker(bind=engine)
            self.session = Session()

        def tearDown(self):
            self.session.close()

        def test_getter_returns_none_if_proxy_value_is_absent(self):
            p = self.Primary()
            self.session.add(p)
            self.assertTrue(p.bar is None)

        def test_getter_returns_proxy_value_if_proxy_value_is_present(self):
            p = self.Primary()
            self.assertTrue(p.bar is None)
            s = self.Secondary(primary=p, bar='something')
            self.session.add(p)
            self.assertEqual(p.bar, 'something')

        def test_setter_assigns_proxy_value(self):
            p = self.Primary()
            s = self.Secondary(primary=p)
            self.session.add(p)
            self.assertTrue(s.bar is None)
            p.bar = 'something'
            self.assertEqual(s.bar, 'something')



# -*- coding: utf-8; -*-

from unittest import TestCase

import pytest

from rattail.vendors import catalogs as mod
from rattail.config import make_config


class TestCatalogParser(TestCase):

    def setUp(self):
        self.config = self.make_config()
        self.app = self.config.get_app()
        self.parser = self.make_parser()

    def make_config(self):
        return make_config([], extend=False)

    def make_parser(self):
        return mod.CatalogParser(self.config)

    def test_key_required(self):

        # someone must define the parser key
        self.assertRaises(NotImplementedError, getattr, self.parser, 'key')

    def test_make_row(self):
        try:
            model = self.app.model
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # make a basic row, it should work
        row = self.parser.make_row()
        self.assertIsInstance(row, model.VendorCatalogBatchRow)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.batch import product as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestProductBatchHandler(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.handler = self.make_handler()

        def make_config(self):
            return make_config([], extend=False)

        def make_handler(self):
            return mod.ProductBatchHandler(self.config)

        def test_make_label_batch(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # prep data
            betty = model.User(username='betty')
            prodbatch = model.ProductBatch(id=1, created_by=betty)
            session.add(prodbatch)
            session.commit()

            # make basic label batch
            lblbatch = self.handler.make_label_batch(prodbatch, betty, id=2)
            self.assertIsNotNone(lblbatch)
            self.assertEqual(lblbatch.id, 2)

        def test_make_pricing_batch(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # prep data
            betty = model.User(username='betty')
            prodbatch = model.ProductBatch(id=1, created_by=betty)
            session.add(prodbatch)
            session.commit()

            # make basic pricing batch
            prcbatch = self.handler.make_pricing_batch(prodbatch, betty, id=2)
            self.assertIsNotNone(prcbatch)
            self.assertEqual(prcbatch.id, 2)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import make_config

try:
    import sqlalchemy as sa
    from rattail.autocomplete import vendors as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestVendorAutocompleter(TestCase):

        def setUp(self):
            self.config = self.make_config()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

        def make_config(self):
            return make_config([], extend=False)

        def make_autocompleter(self):
            return mod.VendorAutocompleter(self.config)

        def test_autocomplete(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)

            # first create some vendors
            acme = model.Vendor(name='Acme Wholesale Foods')
            session.add(acme)
            bigboy = model.Vendor(name='Big Boy Distributors')
            session.add(bigboy)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'd' yields both vendors
            result = self.autocompleter.autocomplete(session, 'd')
            self.assertEqual(len(result), 2)

            # search for 'big' yields just Big Boy
            result = self.autocompleter.autocomplete(session, 'big')
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0]['value'], bigboy.uuid)



# -*- coding: utf-8; -*-
"""rename constraints per wutta

Revision ID: f6e95c74d8db
Revises: 4e7b6d8e71a4
Create Date: 2024-07-14 14:09:27.178397

"""

# revision identifiers, used by Alembic.
revision = 'f6e95c74d8db'
down_revision = '4e7b6d8e71a4'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # uq_role_name
    op.drop_constraint('role_uq_name', 'role', type_='unique')
    op.create_unique_constraint(op.f('uq_role_name'), 'role', ['name'])

    # fk_permission_role_uuid_role
    op.drop_constraint('permission_fk_role', 'permission', type_='foreignkey')
    op.create_foreign_key('fk_permission_role_uuid_role',
                          'permission', 'role',
                          ['role_uuid'], ['uuid'])

    # fk_user_person_uuid_person
    op.drop_constraint('user_fk_person', 'user', type_='foreignkey')
    op.create_foreign_key('fk_user_person_uuid_person',
                          'user', 'person',
                          ['person_uuid'], ['uuid'])

    # uq_user_username
    op.drop_constraint('user_uq_username', 'user', type_='unique')
    op.create_unique_constraint(op.f('uq_user_username'), 'user', ['username'])

    # fk_user_x_role_role_uuid_role
    op.drop_constraint('user_x_role_fk_role', 'user_x_role', type_='foreignkey')
    op.create_foreign_key('fk_user_x_role_role_uuid_role',
                          'user_x_role', 'role',
                          ['role_uuid'], ['uuid'])

    # fk_user_x_role_user_uuid_user
    op.drop_constraint('user_x_role_fk_user', 'user_x_role', type_='foreignkey')
    op.create_foreign_key('fk_user_x_role_user_uuid_user',
                          'user_x_role', 'user',
                          ['user_uuid'], ['uuid'])


def downgrade():

    # fk_user_x_role_user_uuid_user
    op.drop_constraint('fk_user_x_role_user_uuid_user', 'user_x_role', type_='foreignkey')
    op.create_foreign_key('user_x_role_fk_user',
                          'user_x_role', 'user',
                          ['user_uuid'], ['uuid'])

    # fk_user_x_role_role_uuid_role
    op.drop_constraint('fk_user_x_role_role_uuid_role', 'user_x_role', type_='foreignkey')
    op.create_foreign_key('user_x_role_fk_role',
                          'user_x_role', 'role',
                          ['role_uuid'], ['uuid'])

    # uq_user_username
    op.drop_constraint(op.f('uq_user_username'), 'user', type_='unique')
    op.create_unique_constraint('user_uq_username', 'user', ['username'])

    # fk_user_person_uuid_person
    op.drop_constraint('fk_user_person_uuid_person', 'user', type_='foreignkey')
    op.create_foreign_key('user_fk_person',
                          'user', 'person',
                          ['person_uuid'], ['uuid'])

    # fk_permission_role_uuid_role
    op.drop_constraint('fk_permission_role_uuid_role', 'permission', type_='foreignkey')
    op.create_foreign_key('permission_fk_role',
                          'permission', 'role',
                          ['role_uuid'], ['uuid'])

    # uq_role_name
    op.drop_constraint(op.f('uq_role_name'), 'role', type_='unique')
    op.create_unique_constraint('role_uq_name', 'role', ['name'])



# -*- coding: utf-8; -*-

try:
    from rattail.db import model
    from ... import DataTestCase
except ImportError:
    pass
else:

    class TestCustomer(DataTestCase):

        # TODO: this is duplicated in TestPerson
        def test_add_email_address(self):
            customer = model.Customer()
            self.assertEqual(len(customer.emails), 0)
            customer.add_email_address('fred@mailinator.com')
            self.assertEqual(len(customer.emails), 1)
            email = customer.emails[0]
            self.assertEqual(email.type, 'Home')

            customer = model.Customer()
            self.assertEqual(len(customer.emails), 0)
            customer.add_email_address('fred@mailinator.com', type='Work')
            self.assertEqual(len(customer.emails), 1)
            email = customer.emails[0]
            self.assertEqual(email.type, 'Work')

        # TODO: this is duplicated in TestPerson
        def test_add_phone_number(self):
            customer = model.Customer()
            self.assertEqual(len(customer.phones), 0)
            customer.add_phone_number('417-555-1234')
            self.assertEqual(len(customer.phones), 1)
            phone = customer.phones[0]
            self.assertEqual(phone.type, 'Home')

            customer = model.Customer()
            self.assertEqual(len(customer.phones), 0)
            customer.add_phone_number('417-555-1234', type='Work')
            self.assertEqual(len(customer.phones), 1)
            phone = customer.phones[0]
            self.assertEqual(phone.type, 'Work')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for label batches
"""

import csv
import decimal
import logging

import json

from rattail.db.model import LabelBatch
from rattail.batch import BatchHandler
from rattail.csvutil import UnicodeDictReader


log = logging.getLogger(__name__)


class LabelBatchHandler(BatchHandler):
    """
    Handler for Print Labels batches.
    """
    batch_model_class = LabelBatch

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.products_handler = self.app.get_products_handler()

    def setup(self, batch, progress=None):
        self.now = self.app.make_utc()

    setup_populate = setup
    setup_refresh = setup
    setup_clone = setup

    def make_batch(self, session, progress=None, **kwargs):
        """
        Make a new batch, with initial rows if applicable.
        """
        self.skip_first_line = self.config.parse_bool(kwargs.pop('skip_first_line', False))
        self.calc_check_digit = kwargs.pop('calc_check_digit', False)
        if self.calc_check_digit != 'upc':
            self.calc_check_digit = self.config.parse_bool(self.calc_check_digit)
        file_has_options = self.config.parse_bool(kwargs.pop('file_has_options', False))
        batch = super().make_batch(session, progress, **kwargs)
        batch.file_has_options = file_has_options
        return batch

    def auto_executable(self, batch):
        """
        Must return a boolean indicating whether the given bath is eligible for
        "automatic" execution, i.e. immediately after batch is created.
        """
        if batch.filename and '.autoexecute.' in batch.filename:
            return True
        return False

    def populate(self, batch, progress=None):
        """
        Pre-fill batch with row data from handheld batch, etc.
        """
        model = self.model
        session = self.app.get_session(batch)
        if batch.label_profile:
            self.label_profile = batch.label_profile
        else:
            self.label_profile = self.get_label_profile(session)

        if hasattr(batch, 'product_batch') and batch.product_batch:
            self.populate_from_product_batch(batch, progress=progress)
            return

        assert batch.handheld_batch or batch.filename or batch.products
        label_code = self.label_profile.code if self.label_profile else None

        def append(item, i):
            row = self.make_row()
            row.label_code = label_code
            row.label_profile = self.label_profile
            with session.no_autoflush:
                if isinstance(item, model.Product):
                    row.product = item
                    row.label_quantity = 1
                    if batch.static_prices and hasattr(item, '_batch_price'):
                        row.regular_price = item._batch_price
                else: # item is handheld batch row
                    row.product = item.product
                    row.label_quantity = item.units or 1
                    # copy these in case product is null
                    row.item_entry = item.item_entry
                    row.item_id = item.item_id
                    row.upc = item.upc
                    row.brand_name = item.brand_name
                    row.description = item.description
                    row.size = item.size
            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        if batch.handheld_batch:
            data = batch.handheld_batch.active_rows()
        elif batch.filename:
            if batch.file_has_options:
                self.set_options_from_file(batch)
                if batch.label_profile:
                    self.label_profile = batch.label_profile
            data = self.read_products_from_file(batch, progress=progress)
        elif batch.products:
            data = batch.products

        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def populate_from_product_batch(self, batch, progress=None):
        """
        Populate label batch from product batch.
        """
        session = self.app.get_session(batch)
        product_batch = batch.product_batch
        label_code = self.label_profile.code if self.label_profile else None

        def add(prow, i):
            row = self.make_row()
            row.label_code = label_code
            row.label_profile = self.label_profile
            row.label_quantity = 1
            with session.no_autoflush:
                row.product = prow.product
            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        self.progress_loop(add, product_batch.active_rows(), progress,
                           message="Adding initial rows to batch")

    def set_options_from_file(self, batch):
        """
        Set various batch options, if any are present within the data file.
        """
        model = self.model
        path = batch.filepath(self.config)
        with open(path, 'rt') as f:
            options = json.loads(f.readline())
        if 'description' in options and options['description']:
            batch.description = options['description']
        if 'notes' in options and options['notes']:
            batch.notes = options['notes']
        if 'static_prices' in options:
            batch.static_prices = options['static_prices']
        if 'label_code' in options:
            batch.label_code = options['label_code']
            if batch.label_code:
                session = self.app.get_session(batch)
                batch.label_profile = session.query(model.LabelProfile)\
                                             .filter(model.LabelProfile.code == batch.label_code)\
                                             .one()

    def read_products_from_file(self, batch, progress=None):
        """
        Returns list of Product objects based on lookup from CSV file data.

        # TODO: should this actually happen here? vs refresh and just mark product not found?
        """
        path = batch.filepath(self.config)
        with open(path, 'rt') as f:
            if self.skip_first_line:
                f.readline()
                reader = csv.reader(f)
                data = [{'upc': row[0]} for row in reader]
            else:
                fields = None
                if batch.file_has_options:
                    f.readline()
                    reader = csv.reader(f)
                    fields = next(reader)
                    f.seek(0)
                    f.readline()
                    f.readline()
                reader = UnicodeDictReader(f, fieldnames=fields)
                data = list(reader)

        products = []
        session = self.app.get_session(batch)

        def append(entry, i):
            upc = entry['upc'].strip()
            if upc:
                try:
                    upc = self.app.make_gpc(upc, calc_check_digit=self.calc_check_digit)
                except ValueError:
                    pass
                else:
                    product = self.products_handler.locate_product_for_gpc(session, upc)
                    if product:
                        if batch.static_prices and entry['regular_price']:
                            product._batch_price = decimal.Decimal(entry['regular_price'])
                        products.append(product)
                    else:
                        log.warning("product not found: {}".format(upc))

        self.progress_loop(append, data, progress,
                           message="Reading data from CSV file")
        return products

    def get_label_profile(self, session):
        model = self.model
        code = self.config.get('rattail.batch', 'labels.default_code')
        if code:
            return session.query(model.LabelProfile)\
                          .filter(model.LabelProfile.code == code)\
                          .one()
        else:
            return session.query(model.LabelProfile)\
                          .order_by(model.LabelProfile.ordinal)\
                          .first()

    def refresh_row(self, row):
        """
        Inspect a row from the source data and populate additional attributes
        for it, according to what we find in the database.
        """
        if not row.product:
            session = self.app.get_session(row)
            if row.item_entry:
                row.product = self.locate_product_for_entry(session, row.item_entry)
            if not row.product and row.upc:
                row.product = self.products_handler.locate_product_for_gpc(session, row.upc)
            if not row.product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        self.refresh_product_basics(row)

        product = row.product
        category = product.category
        row.category_code = category.code if category else None
        row.category_name = category.name if category else None

        if not row.batch.static_prices:
            regular_price = product.regular_price
            row.regular_price = regular_price.price if regular_price else None
            row.pack_quantity = regular_price.pack_multiple if regular_price else None
            row.pack_price = regular_price.pack_price if regular_price else None

            sale_price = product.sale_price
            if sale_price:
                now = getattr(self, 'now', None) or self.app.make_utc()
                if (sale_price.type == self.enum.PRICE_TYPE_SALE and
                    sale_price.starts and sale_price.starts <= now and
                    sale_price.ends and sale_price.ends >= now):
                    pass            # this is what we want
                else:
                    sale_price = None
            row.sale_price = sale_price.price if sale_price else None
            row.sale_start = sale_price.starts if sale_price else None
            row.sale_stop = sale_price.ends if sale_price else None

            tpr_price = product.tpr_price
            if tpr_price:
                now = getattr(self, 'now', None) or self.app.make_utc()
                if (tpr_price.type == self.enum.PRICE_TYPE_TPR and
                    tpr_price.starts and tpr_price.starts <= now and
                    tpr_price.ends and tpr_price.ends >= now):
                    pass            # this is what we want
                else:
                    tpr_price = None
            row.tpr_price = tpr_price.price if tpr_price else None
            row.tpr_starts = tpr_price.starts if tpr_price else None
            row.tpr_ends = tpr_price.ends if tpr_price else None

            current_price = product.current_price
            if current_price:
                now = getattr(self, 'now', None) or self.app.make_utc()
                if (current_price.type in (self.enum.PRICE_TYPE_SALE,
                                           self.enum.PRICE_TYPE_TPR) and
                    current_price.starts and current_price.starts <= now and
                    current_price.ends and current_price.ends >= now):
                    pass            # this is what we want
                else:
                    current_price = None
            row.current_price = current_price.price if current_price else None
            row.current_starts = current_price.starts if current_price else None
            row.current_ends = current_price.ends if current_price else None

        cost = product.cost
        vendor = cost.vendor if cost else None
        row.vendor_id = vendor.id if vendor else None
        row.vendor_name = vendor.name if vendor else None
        row.vendor_item_code = cost.code if cost else None
        row.case_quantity = cost.case_size if cost else None
        if row.regular_price:
            row.status_code = row.STATUS_OK
        else:
            row.status_code = row.STATUS_REGULAR_PRICE_UNKNOWN

    def quick_entry(self, session, batch, entry):
        """
        Quick entry is assumed to be a UPC scan or similar user input.  If a
        matching product can be found, this will add a new row for the batch;
        otherwise an error is raised.
        """
        product = self.locate_product_for_entry(session, entry)
        if not product:
            raise ValueError("Product not found: {}".format(entry))

        row = self.make_row()
        row.product = product
        self.add_row(batch, row)
        return row

    def get_effective_rows(self, batch):
        # filter out removed rows, and maybe inactive product rows
        rows = batch.active_rows()
        if self.config.getbool('rattail.batch',
                               'labels.exclude_inactive_products',
                               default=False):
            rows = [row for row in rows
                    if row.status_code not in (row.STATUS_PRODUCT_APPEARS_INACTIVE,
                                               row.STATUS_PRODUCT_NOT_FOUND)]
        return rows

    def execute(self, batch, progress=None, **kwargs):
        """
        Print some labels!
        """
        rows = self.get_effective_rows(batch)
        self.print_labels(batch, rows, progress=progress)
        return True

    def print_labels(self, batch, rows, progress=None):
        """
        Print all labels for the given batch.
        """
        label_handler = self.app.get_label_handler()
        profiles = {}

        def organize(row, i):
            profile = row.label_profile
            if not profile:
                return
            if profile.uuid not in profiles:
                profiles[profile.uuid] = profile
                profile.labels = []

            data = row.get_data_dict()

            # TODO: should rename these columns in the schema; for now
            # just copy values in the dict to make other logic happy
            data['sale_starts'] = data['sale_start']
            data['sale_ends'] = data['sale_stop']

            # TODO: not sure what to make of this yet.  printing
            # labels should not require a product i think, but it
            # currently does (cf. CommandFormatter.format_labels).
            # and then again it seems like it would be useful for some
            # label formatters to leverage things from the product
            # which aren't in the batch data.  so for now we just
            # always pass it i guess.
            data['product'] = row.product

            profile.labels.append((data, row.label_quantity))

        self.progress_loop(organize, rows, progress,
                           message="Organizing labels by type")

        # okay now print for real
        for profile in profiles.values():
            printer = label_handler.get_printer(profile)
            printer.print_labels(profile.labels, progress=progress)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Clientele Handler
"""

from collections import OrderedDict
import logging
import warnings

from rattail.app import GenericHandler


log = logging.getLogger(__name__)


class ClienteleHandler(GenericHandler):
    """
    Base class and default implementation for clientele handlers.
    """

    def choice_uses_dropdown(self):
        """
        Returns boolean indicating whether a customer choice should be
        presented to the user via a dropdown (select) element, vs.  an
        autocomplete field.  The latter is the default because
        potentially the customer list can be quite large, so we avoid
        loading them all in the dropdown unless so configured.

        :returns: Boolean; if true then a dropdown should be used;
           otherwise (false) autocomplete is used.
        """
        return self.config.getbool('rattail', 'customers.choice_uses_dropdown',
                                   default=False)

    def ensure_customer(self, person):
        """
        Returns the customer record associated with the given person, creating
        it first if necessary.
        """
        customer = self.get_customer(person)
        if customer:
            return customer

        session = self.app.get_session(person)
        customer = self.make_customer(person)
        return customer

    def get_customer(self, obj):
        """
        Return the Customer associated with the given object, if any.
        """
        model = self.model

        if isinstance(obj, model.Customer):
            return obj

        else:
            person = self.app.get_person(obj)
            if person:
                # TODO: all 3 options below are indeterminate, since it's
                # *possible* for a person to hold multiple accounts
                # etc. but not sure how to fix in a generic way?  maybe
                # just everyone must override as needed
                if person.customer_accounts:
                    return person.customer_accounts[0]
                for shopper in person.customer_shoppers:
                    if shopper.shopper_number == 1:
                        return shopper.customer
                # legacy fallback
                if person.customers:
                    return person.customers[0]

    def get_email_address(self, customer, **kwargs):
        """
        Return the first email address found for the given customer.

        :returns: The email address as string, or ``None``.
        """
        warnings.warn("clientele.get_email_address(customer) is deprecated; please "
                      "use app.get_contact_email_address(customer) instead",
                      DeprecationWarning, stacklevel=2)
        return self.app.get_contact_email_address(customer)

    def get_all_customers(self, session, include_inactive=False, **kwargs):
        """
        Get the full list of customers, e.g. for dropdown choice.

        :param include_inactive: Flag indicating if "inactive"
           customers should be included.  This is false by default, in
           which case only "active" customers are returned.

        :returns: List of
           :class:`~rattail.db.model.customers.Customer` objects.
        """
        import sqlalchemy as sa

        model = self.model
        customers = session.query(model.Customer)\
                           .order_by(model.Customer.name)

        if not include_inactive:
            customers = customers.filter(sa.or_(
                model.Customer.active_in_pos == True,
                model.Customer.active_in_pos == None))

        return customers.all()

    def get_customers_for_account_holder(
            self,
            person,
            **kwargs
    ):
        """
        Return all Customer records for which the given Person is the
        account holder.
        """
        customers = OrderedDict()

        # find customers for which person is account holder
        for customer in person.customer_accounts:
            customers.setdefault(customer.uuid, customer)

        # find customers for which person is primary shopper
        for shopper in person.customer_shoppers:
            if shopper.shopper_number == 1:
                customer = shopper.customer
                customers.setdefault(customer.uuid, customer)

        # nb. legacy
        for customer in person.customers:
            customers.setdefault(customer.uuid, customer)

        return list(customers.values())

    def get_active_shopper(
            self,
            customer,
            **kwargs
    ):
        """
        Return the "active" shopper record for the given customer.

        This should never return multiple shoppers, either one or none.
        """
        for shopper in customer.shoppers:
            if shopper.active:
                return shopper

    def deactivate_shopper(self, shopper, **kwargs):
        """
        Deactivate the given shopper, i.e. make it no longer active
        for the customer account to which it belongs.

        :param shopper: The shopper to be deactivated.

        :param end_date: Optional end date for the deactivation.  If
           not specified, "today" is assumed.
        """
        # declare end date for current shopper history, if applicable
        if 'end_date' in kwargs:
            end_date = kwargs['end_date']
        else:
            end_date = self.app.today()
        if end_date:
            history = shopper.get_current_history()
            if history:
                history.end_date = end_date

        # mark shopper as no longer active
        shopper.active = False

    def shopper_was_active(self, shopper, date, **kwargs):
        """
        Inspect the shopper's history to determine if it was
        considered *active* (for the parent customer account) on the
        given date.

        :param shopper: The shopper to be checked.

        :param date: The date to be checked.

        :returns: Boolean indicating whether shopper was active on
           that date.
        """
        # try to find (all) applicable history for shopper
        applicable = []
        for history in reversed(shopper.history):
            if history.start_date > date:
                # this history began after the date
                continue
            if history.end_date and history.end_date < date:
                # this history ended before the date
                continue
            # okay, got one
            applicable.append(history)

        # if we found at least one "applicable" history record, that
        # means the shopper *was* indeed active on this date
        if applicable:
            # nb. there should only be *one* applicable record,
            # otherwise overlapping history records are implied
            if len(applicable) != 1:
                log.error("found (%s) applicable history records for shopper %s: %s",
                          len(applicable), shopper.uuid, shopper)
            return True

        # if we can't prove shopper was active, we assume they were *inactive*
        return False

    def get_person(self, customer):
        """
        Returns the person associated with the given customer, if there is one.
        """
        warnings.warn("ClienteleHandler.get_person() is deprecated; "
                      "please use AppHandler.get_person() instead")

        return self.app.get_person(customer)

    def make_customer(self, person, **kwargs):
        """
        Create and return a new customer record.
        """
        session = self.app.get_session(person)
        customer = self.model.Customer()
        customer.name = person.display_name
        customer.account_holder = person
        session.add(customer)
        session.flush()
        session.refresh(person)
        return customer

    def locate_customer_for_entry(self, session, entry, **kwargs):
        """
        This method aims to provide sane default logic for locating a
        :class:`~rattail.db.model.customers.Customer` record for the
        given "entry" value.

        The default logic here will try to honor the "configured"
        customer field, and prefer that when attempting the lookup.

        :param session: Reference to current DB session.

        :param entry: Value to use for lookup.  This is most often a
           simple string, but the method can handle a few others.  For
           instance it is common to read values from a spreadsheet,
           and sometimes those come through as integers etc.

        :param lookup_fields: Optional list of fields to use for
           lookup.  The default value is ``['uuid', '_customer_key_']``
           which means to lookup by UUID as well as "customer key"
           field, which is configurable.  You can include any of the
           following in ``lookup_fields``:

           * ``uuid``
           * ``_customer_key_`` - :meth:`locate_customer_for_key`

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance found if there was a match; otherwise ``None``.
        """
        model = self.model
        if not entry:
            return

        # figure out which fields we should match on
        # TODO: let config declare default lookup_fields
        lookup_fields = kwargs.get('lookup_fields', [
            'uuid',
            '_customer_key_',
        ])

        # try to locate customer by uuid before other, more specific key
        if 'uuid' in lookup_fields:
            if isinstance(entry, str):
                customer = session.get(model.Customer, entry)
                if customer:
                    return customer

        lookups = {
            'uuid': None,
            '_customer_key_': self.locate_customer_for_key,
        }

        for field in lookup_fields:
            if field in lookups:
                lookup = lookups[field]
                if lookup:
                    customer = lookup(session, entry, **kwargs)
                    if customer:
                        return customer
            else:
                log.warning("unknown lookup field: %s", field)

    def locate_customer_for_key(self, session, entry, customer_key=None, **kwargs):
        """
        Locate the customer which matches the given key value.

        This is an abstraction layer so calling logic need not care
        which customer key field is configured.  Under the hood this
        will invoke one of:

        * :meth:`locate_customer_for_id`
        * :meth:`locate_customer_for_number`

        This will do a lookup on the customer key field only.  It
        normally checks config to determine which field to use for
        customer key (via
        :meth:`~rattail.app.AppHandler.get_customer_key_field()`), but
        you can override by specifying, e.g.
        ``customer_key='number'``.

        :param session: Current session for Rattail DB.

        :param entry: Key value to use for the lookup.

        :param customer_key: Optional key field to use for the lookup.
           If not specified, will be read from config.

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance if a match was found; otherwise ``None``.
        """
        # prefer caller-provided key over configured key
        if not customer_key:
            customer_key = self.app.get_customer_key_field()

        customer = None

        if customer_key == 'id':
            customer = self.locate_customer_for_id(session, entry, **kwargs)

        elif customer_key == 'number':
            customer = self.locate_customer_for_number(session, entry, **kwargs)

        return customer

    def locate_customer_for_id(self, session, entry, **kwargs):
        """
        Locate the customer which matches the given ID.

        This will do a lookup on the
        :attr:`rattail.db.model.customers.Customer.id` field only.

        Note that instead of calling this method directly, you might
        consider calling :meth:`locate_customer_for_key()` instead.

        :param session: Current session for Rattail DB.

        :param entry: Customer ID value as string.

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance found if there was a match; otherwise ``None``.
        """
        from sqlalchemy import orm

        if not entry:
            return

        # assume entry is string
        entry = str(entry)

        model = self.model
        try:
            return session.query(model.Customer)\
                          .filter(model.Customer.id == entry).one()
        except orm.exc.NoResultFound:
            pass

    def locate_customer_for_number(self, session, entry, **kwargs):
        """
        Locate the customer which matches the given number.

        This will do a lookup on the
        :attr:`rattail.db.model.customers.Customer.number` field only.

        Note that instead of calling this method directly, you might
        consider calling :meth:`locate_customer_for_key()` instead.

        :param session: Current session for Rattail DB.

        :param entry: Customer number, as integer or string.

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance found if there was a match; otherwise ``None``.
        """
        from sqlalchemy import orm

        if not entry:
            return

        # assume entry is integer
        try:
            entry = int(entry)
        except:
            log.debug("cannot coerce to integer: %s", entry)
            return

        model = self.model
        try:
            return session.query(model.Customer)\
                          .filter(model.Customer.number == entry).one()
        except orm.exc.NoResultFound:
            pass

    def search_customers(self, session, entry, **kwargs):
        """
        Perform a customer search across multiple fields, and return
        results as JSON data rows.
        """
        model = self.model
        final_results = []

        # first we'll attempt "lookup" logic..

        lookup_fields = kwargs.get('lookup_fields', [
            '_customer_key_',
        ])

        if lookup_fields:
            customer = self.locate_customer_for_entry(
                session, entry, lookup_fields=lookup_fields)
            if customer:
                final_results.append(customer)

        # then we'll attempt "search" logic..

        search_fields = kwargs.get('search_fields', [
            'name',
            'email_address',
            'phone_number',
        ])

        searches = {
            'name': self.search_customers_for_name,
            'email_address': self.search_customers_for_email_address,
            'phone_number': self.search_customers_for_phone_number,
        }

        for field in search_fields:
            if field in searches:
                search = searches[field]
                if search:
                    customers = search(session, entry, **kwargs)
                    final_results.extend(customers)
            else:
                log.warning("unknown search field: %s", field)

        return [self.normalize_customer(c)
                for c in final_results]

    def search_customers_for_name(self, session, entry, **kwargs):
        model = self.model
        entry = entry.lower()

        customers = session.query(model.Customer)\
                           .filter(model.Customer.name.ilike(f'%{entry}%'))\
                           .all()
        results = customers

        return results

    def search_customers_for_email_address(self, session, entry, **kwargs):
        model = self.model
        entry = entry.lower()

        customers = session.query(model.Customer)\
                           .join(model.CustomerEmailAddress,
                                 model.CustomerEmailAddress.parent_uuid == model.Customer.uuid)\
                           .filter(model.CustomerEmailAddress.address.ilike(f'%{entry}%'))\
                           .all()
        results = customers

        customers = session.query(model.Customer)\
                           .join(model.Person)\
                           .join(model.PersonEmailAddress,
                                 model.PersonEmailAddress.parent_uuid == model.Person.uuid)\
                           .filter(model.PersonEmailAddress.address.ilike(f'%{entry}%'))\
                           .all()
        results.extend(customers)

        return results

    def search_customers_for_phone_number(self, session, entry, **kwargs):
        model = self.model
        entry = entry.lower()

        customers = session.query(model.Customer)\
                           .join(model.CustomerPhoneNumber,
                                 model.CustomerPhoneNumber.parent_uuid == model.Customer.uuid)\
                           .filter(model.CustomerPhoneNumber.number.ilike(f'%{entry}%'))\
                           .all()
        results = customers

        customers = session.query(model.Customer)\
                           .join(model.Person)\
                           .join(model.PersonPhoneNumber,
                                 model.PersonPhoneNumber.parent_uuid == model.Person.uuid)\
                           .filter(model.PersonPhoneNumber.number.ilike(f'%{entry}%'))\
                           .all()
        results.extend(customers)

        return results

    def normalize_customer(self, customer, fields=None, **kwargs):
        """
        Normalize the given customer to a JSON-serializable dict.
        """
        key = self.app.get_customer_key_field()
        return {
            'uuid': customer.uuid,
            '_customer_key_': getattr(customer, key),
            'id': customer.id,
            'number': customer.number,
            'name': customer.name,
            'phone_number': self.app.get_contact_phone_number(customer),
            'email_address': self.app.get_contact_email_address(customer),
            '_str': str(customer),
        }

    def get_customer_info_markdown(self, customer, **kwargs):
        """
        Returns a Markdown string containing pertinent info about a
        given customer account.
        """
        key_field = self.app.get_customer_key_field()
        key_label = self.app.get_customer_key_label()
        phone = self.app.get_contact_phone_number(customer)
        email = self.app.get_contact_email_address(customer)
        return (f"{key_label}: {getattr(customer, key_field)}\n\n"
                f"Name: {customer.name}\n\n"
                f"Phone: {phone or ''}\n\n"
                f"Email: {email or ''}\n\n")

    def get_first_phone(self, customer, **kwargs):
        """
        Return the first available phone record found, either for the
        customer, or its first person.
        """
        phone = customer.first_phone()
        if phone:
            return phone

        person = self.app.get_person(customer)
        if person:
            return person.first_phone()

    def get_first_phone_number(self, customer, **kwargs):
        """
        Return the first available phone number found, either for the
        customer, or its first person.
        """
        phone = self.get_first_phone(customer)
        if phone:
            return phone.number

    def get_first_email(self, customer, invalid=False, **kwargs):
        """
        Return the first available email record found, either for the
        customer, or its first person.
        """
        email = customer.first_email(invalid=invalid)
        if email:
            return email

        person = self.app.get_person(customer)
        if person:
            return person.first_email(invalid=invalid)

    def get_first_email_address(self, customer, invalid=False, **kwargs):
        """
        Return the first available email address found, either for the
        customer, or its first person.
        """
        email = self.get_first_email(customer, invalid=invalid)
        if email:
            return email.address


def get_clientele_handler(config, **kwargs):
    """
    Create and return the configured :class:`ClienteleHandler` instance.
    """
    app = config.get_app()
    spec = config.get('rattail', 'clientele.handler')
    if spec:
        factory = app.load_object(spec)
    else:
        factory = ClienteleHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch

try:
    import sqlalchemy as sa
    from rattail.db import model, Session, SessionBase
    from rattail.importing import rattail as rattail_importing
    from .. import RattailMixin, RattailTestCase
    from . import ImporterTester
except ImportError:
    pass
else:

    class DualRattailMixin(RattailMixin):

        def setup_rattail(self):
            super().setup_rattail()

            if 'host' not in self.config.rattail_engines:
                self.config.rattail_engines['host'] = sa.create_engine('sqlite://')

            self.host_engine = self.config.rattail_engines['host']
            self.config.setdefault('rattail', 'timezone.default', 'America/Chicago')
            self.config.setdefault('rattail.db', 'keys', 'default, host')
            self.config.setdefault('rattail.db', 'host.url', str(self.host_engine.url))
            model = self.get_rattail_model()
            model.Base.metadata.create_all(bind=self.host_engine)
            self.host_session = Session(bind=self.host_engine)

        def teardown_rattail(self):
            super().teardown_rattail()

            self.host_session.close()
            model = self.get_rattail_model()
            model.Base.metadata.drop_all(bind=self.config.rattail_engines['host'])

            if hasattr(self, 'tempio'):
                self.tempio = None


    class DualRattailTestCase(DualRattailMixin, TestCase):
        pass


    class TestFromRattailHandler(RattailTestCase, ImporterTester):
        handler_class = rattail_importing.FromRattailHandler

        def test_make_host_session(self):
            handler = self.make_handler()
            session = handler.make_host_session()
            self.assertIsInstance(session, SessionBase)
            self.assertIs(session.bind, self.config.rattail_engine)


    class TestFromRattailToRattail(DualRattailTestCase, ImporterTester):
        handler_class = rattail_importing.FromRattailToRattailImport
        # extend_config = False

        # def test_host_title(self):
        #     handler = self.make_handler(dbkey='host')
        #     self.assertEqual(handler.host_title, "Rattail (host)")

        # TODO
        def test_default_keys(self):
            handler = self.make_handler()
            handler.get_default_keys()

        def test_make_session(self):
            handler = self.make_handler()
            session = handler.make_session()
            self.assertIsInstance(session, SessionBase)
            self.assertIs(session.bind, self.config.rattail_engine)

        def test_make_host_session(self):

            # invalid dbkey
            handler = self.make_handler(dbkey='other')
            self.assertRaises(KeyError, handler.make_host_session)

            # alternate dbkey
            self.config.rattail_engines['other'] = self.config.rattail_engines['host']
            handler = self.make_handler(dbkey='other')
            session = handler.make_host_session()
            self.assertIsInstance(session, SessionBase)
            self.assertIs(session.bind, self.host_engine)


    class TestFromRattail(DualRattailTestCase):

        def make_importer(self, model_class=None, **kwargs):
            kwargs.setdefault('host_session', self.host_session)
            importer = rattail_importing.FromRattail(self.config, **kwargs)
            if model_class:
                importer.model_class = model_class
            return importer

        def test_host_model_class(self):
            importer = self.make_importer()
            self.assertIsNone(importer.model_class)
            self.assertIsNone(importer.host_model_class)
            importer = self.make_importer(model.Product)
            self.assertIs(importer.host_model_class, model.Product)

        def test_query(self):
            importer = self.make_importer(model.Product)
            importer.query()

        def test_normalize_host_object(self):
            importer = self.make_importer(model.Product)
            product = model.Product()
            with patch.object(importer, 'normalize_local_object') as normalize_local:
                normalize_local.return_value = {}
                data = importer.normalize_host_object(product)
                self.assertEqual(data, {})
                normalize_local.assert_called_once_with(product)
            self.assertEqual(data, importer.normalize_local_object(product))


    class TestAdminUser(DualRattailTestCase):

        importer_class = rattail_importing.AdminUserImporter

        def make_importer(self, **kwargs):
            kwargs.setdefault('config', self.config)
            kwargs.setdefault('session', self.session)
            return self.importer_class(**kwargs)

        def get_admin(self):
            app = self.config.get_app()
            auth = app.get_auth_handler()
            return auth.get_role_administrator(self.session)

        def test_normalize_host_object(self):
            importer = self.make_importer()
            importer.setup()

            user = model.User()
            user.username = 'fred'
            self.session.add(user)
            self.session.flush()

            data = importer.normalize_host_object(user)
            self.assertFalse(data['admin'])

            user.roles.append(self.get_admin())
            self.session.flush()
            data = importer.normalize_host_object(user)
            self.assertTrue(data['admin'])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Handler

See also :doc:`rattail-manual:base/handlers/other/auth`.
"""

import secrets
import warnings

from sqlalchemy import orm
import sqlalchemy_continuum as continuum

from wuttjamaican import auth as base

from rattail.app import MergeMixin


class RattailAuthHandler(base.AuthHandler, MergeMixin):
    """
    Default :term:`auth handler` for Rattail.

    This is a subclass of :class:`wuttjamaican.auth.AuthHandler` but
    adds various methods and logic for Rattail.
    """

    ##############################
    # override methods
    ##############################

    def delete_user(self, user, **kwargs):
        """
        Delete the given user account.  Use with caution!  As this
        generally cannot be undone.

        Default behavior here is of course to delete the account, but
        it also tries to remove the user association from various
        places, in particular the continuum transactions table.

        .. warning::

           Please note that if the user was associated with any
           continuum transactions, the "author" for those transactions
           will be set to null.

        Depending on the DB schema and data present, deleting the user
        may still fail with an error (i.e. if the user is still
        referenced by other tables).
        """
        session = self.app.get_session(user)

        # TODO: once we can move the versioning to wuttjamaican, we
        # can stop overriding this method altogether

        # disassociate user from transactions
        if self.config.versioning_has_been_enabled:
            self.remove_user_from_continuum_transactions(user)

        session.delete(user)

        # TODO: should make sure no callers are expecting this!
        return True

    ##############################
    # extra methods
    ##############################

    def authenticate_user_token(self, session, token):
        """
        Authenticate the given user API token string, and if valid,
        return the corresponding User object.
        """
        model = self.app.model

        try:
            token = session.query(model.UserAPIToken)\
                           .filter(model.UserAPIToken.token_string == token)\
                           .one()
        except orm.exc.NoResultFound:
            pass
        else:
            user = token.user
            if user.active:
                return user

    def get_short_display_name(self, user, **kwargs):
        """
        Returns "short display name" for the user.  This is for
        convenience of mobile view, at least...
        """
        # TODO: this should reference employee.short_name
        employee = self.app.get_employee(user)
        if employee and employee.display_name:
            return employee.display_name

        person = self.app.get_person(user)
        if person:
            if person.first_name and person.last_name:
                return "{} {}.".format(person.first_name, person.last_name[0])
            if person.first_name:
                return person.first_name

        return user.username

    def generate_raw_api_token(self):
        """
        Generate a new *raw* API token string.
        """
        return secrets.token_urlsafe()

    def add_api_token(self, user, description, **kwargs):
        """
        Add a new API token for the user.
        """
        model = self.app.model
        session = self.app.get_session(user)

        # generate raw API token, in the form required for use within
        # the API client
        token_string = self.generate_raw_api_token()

        # create DB record for the token
        token = model.UserAPIToken(
            user=user,
            description=description,
            token_string=token_string)
        session.add(token)

        return token

    def delete_api_token(self, token, **kwargs):
        """
        Delete a new API token for the user.
        """
        session = self.app.get_session(token)
        session.delete(token)

    def get_merge_preview_fields(self, **kwargs):
        """
        Returns a sequence of fields which will be used during a merge
        preview.
        """
        F = self.make_merge_field
        return [
            F('uuid'),
            F('username'),
            F('person_uuid', coalesce=True),
            F('person_name', coalesce=True),
            F('role_count'),    # coalesced manually
            F('active', coalesce=True),
            F('sent_message_count', additive=True),
            F('received_message_count', additive=True),
        ]

    def get_merge_preview_data(self, user, **kwargs):
        return {
            'uuid': user.uuid,
            'username': user.username,
            'person_uuid': user.person_uuid,
            'person_name': user.person.display_name if user.person else None,
            '_roles': user.roles, # needed for final role count
            'role_count': len(user.roles),
            'active': user.active,
            'sent_message_count': len(user.sent_messages),
            'received_message_count': len(user._messages),
        }

    def get_merge_resulting_data(self, removing, keeping, **kwargs):
        result = super().get_merge_resulting_data(removing, keeping, **kwargs)

        # nb. must "manually" coalesce the role count
        result['role_count'] = len(set(removing['_roles'] + keeping['_roles']))

        return result

    def why_not_merge(self, removing, keeping, **kwargs):

        if removing.sent_messages:
            return "Cannot (yet) remove a user who has sent messages"

        if removing._messages:
            return "Cannot (yet) remove a user who has received messages"

        if removing._roles:
            return "Cannot (yet) remove a user who is assigned to roles"

    def merge_update_keeping_object(self, removing, keeping):
        super().merge_update_keeping_object(removing, keeping)
        session = self.app.get_session(keeping)
        model = self.app.model

        # update any notes authored by old user, to reflect new user
        notes = session.query(model.Note)\
                       .filter(model.Note.created_by == removing)\
                       .all()
        for note in notes:
            note.created_by = keeping

    def remove_user_from_continuum_transactions(self, user):
        """
        Remove the given user from all Continuum transactions,
        i.e. all data versioning tables.

        You probably will not need to invoke this directly; it is
        invoked as needed from within :meth:`delete_user()`.

        :param user: A :class:`~rattail.db.model.users.User` instance
           which should be purged from the versioning tables.
        """
        session = self.app.get_session(user)
        model = self.app.model

        # remove the user from any continuum transactions
        # nb. we can use "any" model class here, to obtain Transaction
        Transaction = continuum.transaction_class(model.User)
        transactions = session.query(Transaction)\
                              .filter(Transaction.user_id == user.uuid)\
                              .all()
        for txn in transactions:
            txn.user_id = None

    ##############################
    # deprecated methods
    ##############################

    def cache_permissions(self, *args, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "get_permissions() method instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_permissions(*args, **kwargs)

    def generate_preferred_username(self, *args, **kwargs):
        """ """
        warnings.warn("generate_preferred_username() is deprecated; "
                      "please use make_preferred_username() instead",
                      DeprecationWarning, stacklevel=2)
        return self.make_preferred_username(*args, **kwargs)

    def generate_unique_username(self, session, **kwargs):
        """ """
        warnings.warn("generate_unique_username() is deprecated; "
                      "please use make_unique_username() instead",
                      DeprecationWarning, stacklevel=2)
        return self.make_unique_username(*args, **kwargs)

    def generate_username(self, *args, **kwargs): # pragma: no cover
        """ """
        warnings.warn("method is deprecated, please use "
                      "generate_preferred_username() method instead",
                      DeprecationWarning, stacklevel=2)
        return self.generate_preferred_username(*args, **kwargs)

    def get_email_address(self, user, **kwargs):
        """ """
        warnings.warn("auth.get_email_address(user) is deprecated; please "
                      "use app.get_contact_email_address(user) instead",
                      DeprecationWarning, stacklevel=2)
        return self.app.get_contact_email_address(user)

    # nb. technically the method is not deprecated, just a kwarg
    def get_permissions(self, *args, **kwargs):
        """ """
        if 'include_guest' in kwargs:
            warnings.warn("the include_guest param is deprecated; "
                          "please use include_anonymous instead",
                          DeprecationWarning, stacklevel=2)
            kwargs.setdefault('include_anonymous', kwargs.pop('include_guest'))
        return super().get_permissions(*args, **kwargs)

    # nb. technically the method is not deprecated, just a kwarg
    def has_permission(self, *args, **kwargs):
        """ """
        if 'include_guest' in kwargs:
            warnings.warn("the include_guest param is deprecated; "
                          "please use include_anonymous instead",
                          DeprecationWarning, stacklevel=2)
            kwargs.setdefault('include_anonymous', kwargs.pop('include_guest'))
        return super().has_permission(*args, **kwargs)

    ##############################
    # internal methods
    ##############################

    def _role_is_pertinent(self, role):
        if role.node_type:
            if role.node_type == self.config.node_type():
                return True
            return False
        return True


class AuthHandler(RattailAuthHandler):

    def __init__(self, *args, **kwargs):
        warnings.warn("rattail.auth.AuthHandler is deprecated; "
                      "please use RattailAuthHandler instead",
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-

import os
import shutil
import tempfile
from unittest import TestCase

from rattail import db
from rattail.config import RattailConfig


class TestConfigExtension(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        if db.Session:
            db.Session.configure(bind=None, rattail_config=None)
        shutil.rmtree(self.tempdir)

    def write_file(self, fname, content):
        path = os.path.join(self.tempdir, fname)
        with open(path, 'wt') as f:
            f.write(content)
        return path

    def test_configure_empty(self):
        config = RattailConfig()
        self.assertRaises(AttributeError, getattr, config, 'rattail_engines')
        self.assertRaises(AttributeError, getattr, config, 'rattail_engine')
        if hasattr(db.Session, 'kw'):
            self.assertIsNone(db.Session.kw['bind'])
            self.assertIsNone(db.Session.kw['rattail_config'])

        if db.Session:
            db.ConfigExtension().configure(config)
            self.assertEqual(config.rattail_engines, {})
            self.assertIsNone(config.rattail_engine)
            if hasattr(db.Session, 'kw'):
                self.assertIs(db.Session.kw['rattail_config'], config)

    def test_configure_connections(self):
        default_path = self.write_file('default.sqlite', '')
        default_url = 'sqlite:///{}'.format(default_path)
        host_path = self.write_file('host.sqlite', '')
        host_url = 'sqlite:///{}'.format(host_path)

        config = RattailConfig()
        config.setdefault('rattail.db', 'keys', 'default, host')
        config.setdefault('rattail.db', 'default.url', default_url)
        config.setdefault('rattail.db', 'host.url', host_url)
        db.ConfigExtension().configure(config)
        if db.Session:
            self.assertEqual(len(config.rattail_engines), 2)
            self.assertEqual(str(config.rattail_engines['default'].url), default_url)
            self.assertEqual(str(config.rattail_engines['host'].url), host_url)
            self.assertEqual(str(config.rattail_engine.url), default_url)


try:
    import sqlalchemy as sa
except ImportError:
    pass
else:

    class TestSession(TestCase):

        def test_init_rattail_config(self):
            db.Session.configure(rattail_config=None)
            session = db.Session()
            self.assertIsNone(session.rattail_config)
            session.close()

            config = object()
            session = db.Session(rattail_config=config)
            self.assertIs(session.rattail_config, config)
            session.close()

        def test_init_record_changes(self):
            if hasattr(db.Session, 'kw'):
                self.assertIsNone(db.Session.kw.get('rattail_record_changes'))

            session = db.Session()
            self.assertFalse(session.rattail_record_changes)
            session.close()

            session = db.Session(rattail_record_changes=True)
            self.assertTrue(session.rattail_record_changes)
            session.close()

            engine = sa.create_engine('sqlite://')
            engine.rattail_record_changes = True
            session = db.Session(bind=engine)
            self.assertTrue(session.rattail_record_changes)
            session.close()
